# vert class and overloading experiments
import bpy
# PKHG>NEEDED?
import bmesh
from math import acos, pi, sin, cos, atan, tan
from mathutils import Vector
from bpy_extras.object_utils import AddObjectHelper, object_data_add

# PKHG>DBG change the DBG_info and use extra_DBG_info
DBG_info = {"MeshInfo": False, "StrutMesh": False, "HubMesh": False}


def extra_DBG_info(name="value from DBG_info", info_text="default\n", info_obj=None):
    global DBG_info
    DBG_keys = DBG_info.keys()
    if name in DBG_keys:
        if DBG_info[name]:
            print(info_text, info_obj)

sgn = lambda x: (x > 0) - (x < 0)  # missing signum function in Python


def vefm_add_object(selfobj):
    for i in range(len(selfobj.verts)):
        selfobj.verts[i].index = i
    v = [el.vector for el in selfobj.verts]

    e = [[edge.a.index, edge.b.index] for edge in selfobj.edges]

    if type(selfobj.faces[0]) == type([]):
        # PKHG should be a list of vertices, which have an index
        f = [[v.index for v in face] for face in selfobj.faces]
    else:
        f = [[v.index for v in face.vertices] for face in selfobj.faces]

    m = bpy.data.meshes.new(name=selfobj.name)
    m.from_pydata(v, e, f)
    # useful for development when the mesh may be invalid.
    m.validate(verbose=False)
    object_data_add(bpy.context, m, operator=None)


# extra test phase

class vertex:
    def __init__(self, vec=(0, 0, 0)):  # default x = 0, y = 0, z = 0):
        self.vector = Vector(vec)
        self.length = self.vector.length
        self.index = 0
        self.normal = 0
        self.edges = []
        self.faces = []
        self.boundary = 0

    def findlength(self):
        self.length = self.vector.length

    def normalize(self):
        self.findlength()
        if self.length > 0:
            tmp = 1.0 / self.length
            self.vector = tmp * self.vector
            self.length = 1.0

    def findnormal(self):
        target = []
        if self.faces[:] == []:
            print("vefm vertex L68 pkhg:*****ERROR**** findnormal has no faces")
            return
        for currentface in self.faces:
            target.append(currentface.normal)
        self.normal = average(target).centroid()
        self.normal.findlength()
        if self.length == 0:
            print("******ERROR*** lenght zero in findnormal, j = (0,1,0) replcaced")
            self.normal = vertex((0, 1, 0))
        self.normal.normalize()

    def clockwise(self):  # PKHG self is a vertex
        if self.boundary:
            start = self.boundarystart()
        else:
            start = self.faces[0]

        self.tempedges = []
        self.tempfaces = []
        for i in range(len(self.edges)):

            self.tempfaces.append(start)
            for corner in start.corners:
                if corner[0] is not self:
                    pass
                elif corner[0] is self:
                    self.tempedges.append(corner[1])
                    nextedge = corner[2]
            for facey in nextedge.faces:
                if facey is not start:
                    start = facey
                    break
        self.edges = self.tempedges
        self.faces = self.tempfaces

    def boundarystart(self):

        pass

    def __add__(self, other):
        if isinstance(other, Vector):
            tmp = self.vector + other
        else:
            tmp = self.vector + other.vector
        return vertex(tmp)

    def __sub__(self, other):
        if isinstance(other, Vector):
            tmp = self.vector - other
        else:
            tmp = self.vector - other.vector
        return vertex(tmp)

    def __mul__(self, other):
        tmp = self.vector * other
        return vertex(tmp)

    def __truediv__(self, other):
        denom = 1.0 / other
        tmp = self.vector * denom
        return (tmp)

    def negative(self):
        return vertex(-self.vector)


class crossp:
    # Takes in two vertices(vectors), returns the cross product.
    def __init__(self, v1, v2):
        self.v1 = v1
        self.v2 = v2

    def docrossproduct(self):
        tmp = self.v1.vector.cross(self.v2.vector)
        return vertex(tmp)


class average:
    # Takes a list of vertices and returns the average. If two verts are passed, returns midpoint.
    def __init__(self, vertlist):
        self.vertlist = vertlist

    def centroid(self):
        tmp = Vector()
        # PKHG avoid emptylist problems
        divisor = 1.0
        nr_vertices = len(self.vertlist)
        if nr_vertices > 1:
            divisor = 1.0 / len(self.vertlist)
        elif nr_vertices == 0:
            print("\n***WARNING*** empty list in vefm_271.centroid! L158")
        for vert in self.vertlist:
            tmp = tmp + vert.vector
        tmp = tmp * divisor
        return vertex(tmp)


class edge:
    def __init__(self, a=0, b=0):
        self.a = a
        self.b = b
        self.index = 0
        self.normal = 0
        self.cross = 0
        self.unit = 0
        self.faces = []
        self.vect = 0   # PKHG becomes  b - a
        self.vectb = 0  # PKHG becomes  a - b
        self.boundary = 0
        self.findvect()
        self.findlength()

    def findvect(self):
        self.vect = self.b - self.a
        self.vectb = self.a - self.b

    def findlength(self):
        self.vect.findlength()
        self.vectb.length = self.vect.length

    def findnormal(self):
        if self.boundary:
            self.normal = self.faces[0].normal    # average([self.a, self.b]).centroid()
        else:
            self.normal = average([self.faces[0].normal, self.faces[1].normal]).centroid()
        self.normal.normalize()


class face:
    def __init__(self, vertices=[]):
        # PKHG ok good for tri's at least
        self.vertices = vertices    # List of vertex instances
        self.edges = []             # Will be filled with the sides of the face
        self.boundary = 0           # When set will have bool and id of edge concerned
        self.normal = 0             # Face normal found through cross product
        self.corners = []
        self.spokes = []            # Vectors of the bisecting angles from each corner to the centre + dotproduct

        self.index = 0

    # dotproduct is misleading name, it is the hook between two vectors!
    def dotproduct(self, v1, v2):
        v1.findlength()
        v2.findlength()
        if v1.length == 0 or v2.length == 0:
            print("\nPKHG warning, ===== vefm_271 dotproduct L212 ======"
                  " at least one zero vector 0 used")
            return 0
        dot = v1.vector.dot(v2.vector)
        costheta = dot / (v1.length * v2.length)
        tmp = acos(costheta)
        return tmp

    def orderedges(self):
        temp = []
        finish = len(self.vertices)
        for i in range(finish):
            current = self.vertices[i]
            if i == finish - 1:
                next = self.vertices[0]
            else:
                next = self.vertices[i + 1]
            for edge in face.edges:
                if edge.a == current and edge.b == next:
                    face.clockw.append(edge.vect)
                    face.aclockw.append(edge.vectb)
                    temp.append(edge)
                if edge.b == current and edge.a == next:
                    face.clockw.append(edge.vectb)
                    face.aclockw.append(edge.vect)
                    temp.append(edge)
            for edge in face.edges:
                if edge.a == current and edge.b == next:
                    face.clockw.append(edge.vect)
                    face.aclockw.append(edge.vectb)
                    temp.append(edge)
                if edge.b == current and edge.a == next:
                    face.clockw.append(edge.vectb)
                    face.aclockw.append(edge.vect)
                    temp.append(edge)
            face.vertices = temp

    def docorners(self):
        # This function identifies and stores the vectors coming from each vertex
        # allowing easier calculation of cross and dot products.
        finish = len(self.vertices)

        for i in range(finish):
            current = self.vertices[i]
            if i == finish - 1:
                next = self.vertices[0]
            else:
                next = self.vertices[i + 1]
            if i == 0:
                previous = self.vertices[-1]
            else:
                previous = self.vertices[i - 1]
            corner = [current]  # PKHG new for each vertex = current
            # corner = current
            rightedge = None
            leftedge = None
            teller = -1
            for edge in self.edges:
                if finish == 3 and len(self.edges) == 2 and i == 2:
                    return
                teller += 1
                # next and previous are vertex with respect to ith vertex
                if edge.a is current or edge.b is current:    # does this edge contain our current vert
                    if edge.a is current:
                        if edge.b is next:
                            rightedge = edge
                            rightvect = edge.vect
                        if edge.b is previous:
                            leftedge = edge
                            leftvect = edge.vect
                    elif edge.b is current:
                        if edge.a is next:
                            rightedge = edge
                            rightvect = edge.vectb
                        if edge.a is previous:
                            leftedge = edge
                            leftvect = edge.vectb
            corner.append(rightedge)
            corner.append(leftedge)
            if rightedge and leftedge:

                dotty = self.dotproduct(rightvect, leftvect)
                corner.append(dotty)
            self.corners.append(corner)

    def findnormal(self):
        one = self.corners[1][2]
        two = self.corners[1][1]
        if one.a is self.corners[1][0]:
            one = one.vect
        elif one.b is self.corners[1][0]:
            one = one.vectb
        if two.a is self.corners[1][0]:
            two = two.vect
        elif two.b is self.corners[1][0]:
            two = two.vectb
        self.normal = crossp(one, two).docrossproduct()
        self.normal.findlength()
        self.normal.normalize()

    def dospokes(self):
        for corner in self.corners:
            vert = corner[0]
            right = corner[1]
            left = corner[2]
            if right.a is vert:
                one = vertex(right.vect.vector)
            elif right.b is vert:
                one = vertex(right.vectb.vector)
            if left.a is vert:
                two = vertex(left.vect.vector)
            elif left.b is vert:
                two = vertex(left.vectb.vector)

            one.normalize()
            two.normalize()
            spoke = one + two
            spoke.normalize()
            self.spokes.append(spoke)

    def artspokes(self):
        centre = average(self.vertices).centroid()
        for point in self.vertices:
            newedge = edge(point, centre)
            self.spokes.append(newedge)


class mesh:
    def __init__(self, name="GD_mesh"):
        self.name = name
        self.verts = []
        self.edges = []
        self.faces = []
        self.edgeflag = 0
        self.faceflag = 0
        self.vertexflag = 0
        self.vertedgeflag = 0
        self.vertfaceflag = 0
        self.faceedgeflag = 0
        self.boundaryflag = 0
        self.vertnormalflag = 0
        self.edgenormalflag = 0
        self.facenormalflag = 0
        self.a45 = pi * 0.25
        self.a90 = pi * 0.5
        self.a180 = pi
        self.a270 = pi * 1.5
        self.a360 = pi * 2

    def power(self, a, b):  # Returns a power, including negative numbers
        result = sgn(a) * (abs(a) ** b)
        return result

    def sign(self, d):      # Works out the sign of a number.
        return sgn(d)

    def ellipsecomp(self, efactor, theta):
        if theta == self.a90:
            result = self.a90
        elif theta == self.a180:
            result = self.a180
        elif theta == self.a270:
            result = self.a270
        elif theta == self.a360:
            result = 0.0
        else:
            result = atan(tan(theta) / efactor ** 0.5)
            if result < 0.0:
                if theta > self.a180:
                    result = result + self.a180
                elif theta < self.a180:
                    result = result + self.a180

            if result > 0.0:
                if theta > self.a180:
                    result = result + self.a180
                elif theta < self.a180:
                    result = result
        return result

    def connectivity(self):
        self.dovertedge()
        self.dovertface()
        self.dofaceedge()
        self.boundary()

    def superell(self, n1, uv, turn):
        t1 = sin(uv + turn)
        t1 = abs(t1)
        t1 = t1 ** n1
        t2 = cos(uv + turn)
        t2 = abs(t2)
        t2 = t2 ** n1
        r = self.power(1.0 / (t1 + t2), (1.0 / n1))
        return r

    def superform(self, m, n1, n2, n3, uv, a, b, twist):
        t1 = cos(m * (uv + twist) * .25) * a
        t1 = abs(t1)
        t1 = t1 ** n2
        t2 = sin(m * (uv + twist) * .25) * b
        t2 = abs(t2)
        t2 = t2 ** n3
        r = self.power(1.0 / (t1 + t2), n1)
        return r

    def dovertedge(self):
        if not self.vertedgeflag:
            for vert in self.verts:
                vert.edges = []
            for currentedge in self.edges:
                currentedge.a.edges.append(currentedge)
                currentedge.b.edges.append(currentedge)
        self.vertedgeflag = 1

    def dovertface(self):
        if not self.vertfaceflag:
            for vert in self.verts:
                vert.faces = []
            for face in self.faces:
                for vert in face.vertices:
                    vert.faces.append(face)
        self.vertfaceflag = 1

    def dofaceedge(self):
        self.dovertedge()    # just in case they haven't been done
        self.dovertface()
        if not self.faceedgeflag:
            for edge in self.edges:
                edge.faces = []
            for face in self.faces:
                face.edges = []
            for face in self.faces:
                finish = len(face.vertices)
                for i in range(finish):
                    current = face.vertices[i]
                    if i == finish - 1:
                        next = face.vertices[0]
                    else:
                        next = face.vertices[i + 1]
                    for edge in current.edges:
                        if edge.a is current or edge.b is current:
                            if edge.b is next or edge.a is next:
                                edge.faces.append(face)
                                face.edges.append(edge)
        self.faceedgeflag = 1

    def boundary(self):
        if not self.boundaryflag:
            for edge in self.edges:
                if len(edge.faces) < 2:
                    edge.boundary = 1
                    edge.faces[0].boundary = 1
                    edge.a.boundary = 1
                    edge.b.boundary = 1

    # The functions below turn the basic triangular faces into
    # hexagonal faces, creating the buckyball effect.
    # PKHG seems to work only for meshes with tri's ;-)
    def hexify(self):
        self.hexverts = []
        self.hexedges = []
        self.hexfaces = []
        # PKHG renumbering the index of the verts
        for i in range(len(self.verts)):
            self.verts[i].index = i
        # PKHG renumbering the index of the edges
        for i in range(len(self.edges)):
            self.edges[i].index = i

        self.connectivity()
        hexvert_counter = 0
        for edge in self.edges:

            self.hexshorten(edge, hexvert_counter)
            hexvert_counter += 2  # PKHG two new vertices done

        for face in self.faces:
            self.makehexfaces(face)

        for vert in self.verts:
            vert.clockwise()
            self.hexvertface(vert)
        self.verts = self.hexverts
        self.edges = self.hexedges
        self.faces = self.hexfaces
        self.vertedgeflag = 0
        self.vertfaceflag = 0
        self.faceedgeflag = 0

    def hexshorten(self, currentedge, hexvert_counter):
        third = vertex(currentedge.vect / 3.0)
        newvert1 = vertex(currentedge.a.vector)
        newvert2 = vertex(currentedge.b.vector)
        newvert1 = newvert1 + third
        newvert1.index = hexvert_counter
        newvert2 = newvert2 - third
        newvert2.index = hexvert_counter + 1  # PKHG caller adjusts +=2
        newedge = edge(newvert1, newvert2)
        newedge.index = currentedge.index
        self.hexverts.append(newvert1)
        self.hexverts.append(newvert2)
        self.hexedges.append(newedge)

    def makehexfaces(self, currentface):
        vertices = []
        currentface.docorners()
        for corner in currentface.corners:
            vert = corner[0]
            rightedge = corner[1]
            leftedge = corner[2]
            lid = leftedge.index
            rid = rightedge.index

            if leftedge.a is vert:
                vertices.append(self.hexedges[lid].a)
            elif leftedge.b is vert:
                vertices.append(self.hexedges[lid].b)

            if rightedge.a is vert:
                vertices.append(self.hexedges[rid].a)
            elif rightedge.b is vert:
                vertices.append(self.hexedges[rid].b)

        newface = face(vertices)
        newedge1 = edge(vertices[0], vertices[1])
        newedge2 = edge(vertices[2], vertices[3])
        newedge3 = edge(vertices[4], vertices[5])
        self.hexfaces.append(newface)
        self.hexedges.append(newedge1)
        self.hexedges.append(newedge2)
        self.hexedges.append(newedge3)

    def hexvertface(self, vert):
        vertices = []
        for edge in vert.edges:
            eid = edge.index
            if edge.a is vert:
                vertices.append(self.hexedges[eid].a)
            elif edge.b is vert:
                vertices.append(self.hexedges[eid].b)
        newface = face(vertices)
        self.hexfaces.append(newface)

    def starify(self):
        self.starverts = []
        self.staredges = []
        self.starfaces = []
        for i in range(len(self.verts)):
            self.verts[i].index = i
        for i in range(len(self.edges)):
            self.edges[i].index = i
        self.connectivity()
        star_vert_counter = 0
        for currentedge in self.edges:
            newvert = average([currentedge.a, currentedge.b]).centroid()
            newvert.index = star_vert_counter
            star_vert_counter += 1
            self.starverts.append(newvert)
        star_face_counter = 0
        star_edge_counter = 0
        for currentface in self.faces:
            currentface.docorners()
            vertices = []
            for corner in currentface.corners:
                vert = self.starverts[corner[1].index]
                vertices.append(vert)
            newface = face(vertices)
            newface.index = star_face_counter
            star_face_counter += 1
            newedge1 = edge(vertices[0], vertices[1])
            newedge1.index = star_edge_counter
            newedge2 = edge(vertices[1], vertices[2])
            newedge2.index = star_edge_counter + 1
            newedge3 = edge(vertices[2], vertices[0])
            newedge3.index = star_edge_counter + 2
            star_edge_counter += 3
            self.starfaces.append(newface)
            self.staredges.append(newedge1)
            self.staredges.append(newedge2)
            self.staredges.append(newedge3)
        for vert in self.verts:
            vertices = []
            vert.clockwise()
            for currentedge in vert.edges:
                eid = currentedge.index
                vertices.append(self.starverts[eid])
            newface = face(vertices)
            newface.index = star_face_counter
            star_face_counter += 1
            self.starfaces.append(newface)
        self.verts = self.starverts
        self.edges = self.staredges
        self.faces = self.starfaces
        self.vertedgeflag = 0
        self.vertfaceflag = 0
        self.faceedgeflag = 0

    def class2(self):
        self.class2verts = []
        self.class2edges = []
        self.class2faces = []

        newvertstart = len(self.verts)
        newedgestart = len(self.edges)
        counter_verts = len(self.verts)
        for i in range(counter_verts):
            self.verts[i].index = i
        for i in range(len(self.edges)):
            self.edges[i].index = i
        for i in range(len(self.faces)):
            self.faces[i].index = i
        self.connectivity()
        for currentface in self.faces:
            currentface.docorners()
            newvert = average(currentface.vertices).centroid()
            newvert.index = counter_verts

            counter_verts += 1
            self.verts.append(newvert)
            newedge1 = edge(currentface.vertices[0], newvert)
            newedge2 = edge(currentface.vertices[1], newvert)
            newedge3 = edge(currentface.vertices[2], newvert)

            self.edges.append(newedge1)
            self.edges.append(newedge2)
            self.edges.append(newedge3)
        for currentedge in range(newedgestart):
            self.edges[currentedge].a = self.verts[self.edges[currentedge].faces[0].index + newvertstart]
            self.edges[currentedge].b = self.verts[self.edges[currentedge].faces[1].index + newvertstart]
            self.edges[currentedge].findvect()

        for currentvert in range(newvertstart):
            vert = self.verts[currentvert]
            vertices = []
            vert.clockwise()
            for currentface in vert.faces:
                eid = currentface.index
                vertices.append(self.verts[newvertstart + eid])

            for i in range(len(vertices)):
                if i == len(vertices) - 1:
                    next = vertices[0]
                else:
                    next = vertices[i + 1]

                newface = face([vert, vertices[i], next])
                self.class2faces.append(newface)

        self.faces = self.class2faces
        self.vertedgeflag = 0
        self.vertfaceflag = 0
        self.faceedgeflag = 0

    def dual(self):
        self.dualverts = []

        self.dualfaces = []

        counter_verts = len(self.verts)
        for i in range(counter_verts):
            self.verts[i].index = i
        for i in range(len(self.edges)):
            self.edges[i].index = i
        for i in range(len(self.faces)):
            self.faces[i].index = i
        self.connectivity()
        counter_verts = 0
        for currentface in self.faces:
            currentface.docorners()
            newvert = average(currentface.vertices).centroid()
            newvert.index = counter_verts  # PKHG needed in >= 2.59
            counter_verts += 1
            self.dualverts.append(newvert)
        for vert in self.verts:
            vertices = []
            vert.clockwise()
            for currentface in vert.faces:
                eid = currentface.index
                vertices.append(self.dualverts[eid])
            newface = face(vertices)
            self.dualfaces.append(newface)
        for currentedge in self.edges:
            currentedge.a = self.dualverts[currentedge.faces[0].index]
            currentedge.b = self.dualverts[currentedge.faces[1].index]
        self.verts = self.dualverts

        self.faces = self.dualfaces
        self.vertedgeflag = 0
        self.vertfaceflag = 0
        self.faceedgeflag = 0


class facetype(mesh):
    def __init__(self, basegeodesic, parameters, width, height, relative):
        mesh.__init__(self)
        self.detatch = parameters[0]
        self.endtype = parameters[1]
        self.coords = parameters[2]
        self.base = basegeodesic
        self.relative = relative
        self.width = width

        if not self.relative:
            newwidth = self.findrelative()
            self.width = width * newwidth
        self.height = height
        self.base.connectivity()
        for coord in self.coords:
            coord[0] = coord[0] * self.width
            coord[1] = coord[1] * self.height
        if not self.base.facenormalflag:
            for currentface in self.base.faces:

                currentface.docorners()
                currentface.findnormal()

            self.base.facenormalflag = 1
        if self.endtype == 4 and not self.base.vertnormalflag:
            for currentvert in self.base.verts:
                currentvert.findnormal()
            self.base.vertnormalflag = 1
        self.createfaces()

    def findrelative(self):
        centre = average(self.base.faces[0].vertices).centroid()
        edgelist = []
        for point in self.base.faces[0].vertices:
            newedge = edge(centre, point)
            edgelist.append(newedge)
        length = 0
        for edg in edgelist:
            extra = edg.vect.length
            length = length + extra

        length = length / len(edgelist)

        return length

    def createfaces(self):
        if not self.detatch:
            for point in self.base.verts:
                self.verts.append(point)
        if self.endtype == 4:
            self.createghostverts()
        for currentface in self.base.faces:
            self.doface(currentface)

    def createghostverts(self):
        self.ghoststart = len(self.verts)
        for vert in self.base.verts:
            newvert = vert + (vert.normal * self.coords[-1][1])
            self.verts.append(newvert)

    def doface(self, candidate):
        grid = []
        candidate.dospokes()

        if not self.detatch:
            line = []
            for vert in candidate.vertices:
                line.append(vert)
            grid.append(line)
        else:
            line = []
            for point in candidate.vertices:
                newvert = vertex(point.vector)
                self.verts.append(newvert)
                line.append(newvert)
            grid.append(line)
        finish = len(self.coords)
        if self.endtype == 1 or self.endtype == 4:
            finish = finish - 1
        for i in range(finish):
            up = candidate.normal * self.coords[i][1]
            line = []
            for j in range(len(candidate.vertices)):
                dotfac = candidate.corners[j][3] * 0.5
                vec = (candidate.spokes[j] * (self.coords[i][0] / sin(dotfac)))

                newvert = candidate.vertices[j] + vec + up
                line.append(newvert)
                self.verts.append(newvert)
            grid.append(line)
        if self.endtype == 4:
            line = []
            for i in range(len(candidate.vertices)):
                vert = self.verts[candidate.vertices[i].index + self.ghoststart]
                line.append(vert)

            grid.append(line)
        for line in grid:
            line.append(line[0])
        if self.endtype == 3:
            grid.append(grid[0])
        for i in range(len(grid) - 1):
            for j in range(len(grid[i]) - 1):
                one = grid[i][j]
                two = grid[i][j + 1]
                three = grid[i + 1][j + 1]
                four = grid[i + 1][j]
                newface = face([one, two, three, four])
                self.faces.append(newface)
        if self.endtype == 2:
            finalfaceverts = grid[-1]
            newface = face(finalfaceverts[:-1])
            self.faces.append(newface)
        if self.endtype == 1:
            lastvert = average(candidate.vertices).centroid()
            up = candidate.normal * self.coords[-1][1]
            newvert = lastvert + up
            self.verts.append(newvert)
            ring = grid[-1]
            for i in range(len(ring) - 1):
                newface = face([newvert, ring[i], ring[i + 1]])
                self.faces.append(newface)


class importmesh(mesh):
    def __init__(self, meshname, breakquadflag):
        mesh.__init__(self)

        obj = bpy.data.objects[meshname]
        bpy.context.scene.objects.active = obj
        obj.select = True
        impmesh = None
        if not breakquadflag:
            bpy.ops.object.mode_set(mode='EDIT')
            impmesh = bmesh.new()           # create an empty BMesh
            impmesh.from_mesh(obj.data)     # fill it in from a Mesh
            bpy.ops.object.mode_set(mode='OBJECT')

        if breakquadflag:
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.quads_convert_to_tris()
            impmesh = bmesh.new()         # create an empty BMesh
            impmesh.from_mesh(obj.data)   # fill it in from a Mesh
            bpy.ops.object.mode_set(mode='OBJECT')

        for v in impmesh.verts:
            vert = vertex(v.co)
            vert.index = v.index
            self.verts.append(vert)
            # PKHG verts is now a list of vertex, so to say a copy of the Vectors

        # PKHG edges
        for e in impmesh.edges:
            tmp = []
            for vert in e.verts:
                a = self.verts[vert.index]
                tmp.append(a)
            newedge = edge(tmp[0], tmp[1])
            newedge.index = e.index
            self.edges.append(newedge)
        # PKHG faces
        extra_DBG_info("MeshInfo", "vefm L868 the mesh impmesh", impmesh.faces[:])

        for f in impmesh.faces:
            temp = []
            for vert in f.verts:            # PKHG a list! of indices ??? PKHG>???
                a = self.verts[vert.index]  # PKHG verts contains already vertex objects
                temp.append(a)
            newface = face(temp)
            newface.index = f.index  # indexcount
            self.faces.append(newface)
        self.dovertedge()
        self.dovertface()
        self.temp = []

        for i in range(len(self.verts)):
            self.temp.append([])
            self.verts[i].index = i
        for i in range(len(self.verts)):
            target = self.surroundingverts(self.verts[i])
            for j in range(len(target)):                    # go through those verts
                temptarg = self.temp[target[j].index]
                flag = 0                                    # set a flag up

                for k in range(len(temptarg)):              # go through temp list for each of those verts

                    if temptarg[k] == i:                    # if we find a match to the current vert...
                        flag = 1                            # raise the flag

                if flag == 0:                               # if there is no flag after all that...
                    self.temp[target[j].index].append(i)    # add current vert to temp list of this surrounding vert
                    self.temp[i].append(target[j].index)    # add this surrounding vert to the current temp list
                    newedge = edge(self.verts[i], self.verts[target[j].index])
                    self.edges.append(newedge)              # add the newly found edge to the edges list

        for edg in self.edges:
            edg.findvect()
        self.vertedgeflag = 0
        self.vertedgeflag = 0
        self.connectivity()

    def surroundingverts(self, vert):
        ''' Find the verts surrounding vert'''
        surround = []                       # list to be filled and returned
        for faces in vert.faces:            # loop through faces attached to vert
            finish = len(faces.vertices)
            for i in range(finish):
                if i == finish - 1:
                    next = faces.vertices[0]
                else:
                    next = faces.vertices[i + 1]
                if vert == faces.vertices[i]:
                    surround.append(next)
        return surround

    def breakquad(self, quad_face):
        ''' turn quads into triangles'''
        distance1 = quad_face.vertices[0] - quad_face.vertices[2]
        distance2 = quad_face.vertices[1] - quad_face.vertices[3]
        distance1.findlength()
        distance2.findlength()
        if abs(distance1.length) < abs(distance2.length):
            self.faces[quad_face.index] = face([quad_face.vertices[0], quad_face.vertices[1], quad_face.vertices[2]])
            self.faces.append(face([quad_face.vertices[0], quad_face.vertices[2], quad_face.vertices[3]]))
        else:
            self.faces[quad_face.index] = face([quad_face.vertices[0], quad_face.vertices[1], quad_face.vertices[3]])
            self.faces.append(face([quad_face.vertices[1], quad_face.vertices[2], quad_face.vertices[3]]))


class strut(mesh):
    def __init__(self, base, struttype, width, height, length, widthtog, heighttog,
                 lengthtog, meshname, stretchflag, lift):

        extra_DBG_info(name="StrutMesh", info_text="vefm L940\nstrut called: ",
                        info_obj=[base, struttype, width, height, length, widthtog,
                        heighttog, lengthtog, meshname, stretchflag, lift])
        mesh.__init__(self)
        # put in strut prep stuff here
        if struttype is None:
            return
        total = 0
        divvy = len(base.faces[0].edges)
        for lengf in base.faces[0].edges:
            lengf.vect.findlength()
            total = total + lengf.vect.length
        yardstick = total / divvy
        if widthtog:
            self.width = width
        else:
            self.width = width * yardstick
        if heighttog:
            self.height = height
        else:
            self.height = height * yardstick
        if lengthtog:
            self.shrink = length
        else:
            self.shrink = length * yardstick
        if not base.facenormalflag:
            for currentface in base.faces:
                currentface.docorners()
                currentface.findnormal()
            base.facenormalflag = 1
        for edj in base.edges:
            edj.findnormal()
            side = edge(edj.a, edj.b)
            edj.unit = side.vect
            edj.unit.normalize()
            edj.cross = crossp(edj.normal, edj.unit).docrossproduct()
        template = importmesh(meshname, 0)
        maxx = 0
        minx = 0
        for vert in template.verts:
            if vert.vector.x > maxx:
                maxx = vert.vector.x
            if vert.vector.x < minx:
                minx = vert.vector.x
        for edj in base.edges:
            start = len(self.verts)
            centre = average([edj.a, edj.b]).centroid()
            split = edj.vect.length / 2
            # PKHG no division by zero!!
            tmp = 1.0
            if maxx != minx:
                tmp = 1.0 / (maxx - minx)
            dubbl = edj.vect.length * tmp
            # PKHG end no division by zero!!
            diffplus = split - maxx
            diffminus = -split - minx
            for point in template.verts:
                ay = (edj.normal * point.vector.z * self.height) + (edj.normal * lift)
                ce = edj.cross * point.vector.y * self.width

                if stretchflag:
                    be = edj.unit * self.shrink * dubbl * point.vector.x
                else:
                    if point.vector.x > 0.0:
                        be = edj.unit * self.shrink * (point.vector.x + diffplus)
                    elif point.vector.x < 0.0:
                        be = edj.unit * self.shrink * (point.vector.x + diffminus)
                    elif point.vector.x == 0.0:
                        be = edj.unit * self.shrink * point.vector.x
                de = ay + be + ce
                newvert = centre + de
                self.verts.append(newvert)
            for edjy in template.edges:
                one = edjy.a.index + start
                two = edjy.b.index + start
                newedge = edge(self.verts[one], self.verts[two])
                self.edges.append(newedge)
            for facey in template.faces:
                faceverts = []
                for verty in facey.vertices:
                    index = verty.index + start
                    faceverts.append(self.verts[index])
                newface = face(faceverts)
                self.faces.append(newface)
        self.vertedgeflag = 0
        self.vertedgeflag = 0
        self.connectivity()


class hub(mesh):
    def __init__(self, base, hubtype, width, height, length,
                 widthtog, heighttog, lengthtog, meshname):
        mesh.__init__(self)
        self.width = 1.0
        self.height = 1.0
        self.shrink = 1.0
        # put in strut prep stuff here
        extra_DBG_info("vefm L1037 HubMesh", "base is ", str(dir(base)) + "\n and meshname = " + meshname)
        if hubtype is None:
            return
        total = 0
        divvy = len(base.faces[0].edges)
        for lengf in base.verts[0].edges:
            lengf.vect.findlength()
            total = total + lengf.vect.length
        yardstick = total / divvy
        if widthtog:
            self.width = width
        else:
            self.width = width * yardstick
        if heighttog:
            self.height = height
        else:
            self.height = height * yardstick
        if lengthtog:
            self.shrink = length
        else:
            self.shrink = length * yardstick

        if not base.facenormalflag:
            for currentface in base.faces:
                currentface.docorners()
                currentface.findnormal()
            base.facenormalflag = 1

        for apex in base.verts:
            apex.findnormal()
            side = edge(apex.edges[0].a, apex.edges[0].b)
            apex.unit = side.vect  # PKHG is Vector: b - a
            apex.unit.normalize()
            apex.cross = crossp(apex.normal, apex.unit).docrossproduct()
            apex.unit = crossp(apex.cross, apex.normal).docrossproduct()

        template = importmesh(meshname, 0)
        for apex in base.verts:
            start = len(self.verts)
            centre = apex
            for point in template.verts:
                ay = apex.normal * point.vector.z * self.height
                ce = apex.cross * point.vector.y * self.width
                be = apex.unit * point.vector.x * self.shrink
                de = ay + be + ce
                newvert = centre + de
                self.verts.append(newvert)
            for edjy in template.edges:
                one = edjy.a.index + start
                two = edjy.b.index + start
                newedge = edge(self.verts[one], self.verts[two])
                self.edges.append(newedge)
            for facey in template.faces:
                faceverts = []
                for verty in facey.vertices:
                    index = verty.index + start
                    faceverts.append(self.verts[index])
                newface = face(faceverts)
                self.faces.append(newface)
        self.vertedgeflag = 0
        self.vertedgeflag = 0
        self.connectivity()


# ???PKHG TODO Nmesh used yet wrong!
def finalfill(source, target):
    if source == target:  # PKHG: otherewise >infinite< loop
        print("\n***WARNING*** vefm_271.finalfill L1104 source == target empty mesh used")
        target = mesh()
    # PKHG_??? maybe renumverting and checkkin faces wiht >=4 5 vertices?
    count = 0

    for point in source.verts:
        newvert = vertex(point.vector)
        newvert.index = count
        target.verts.append(newvert)
        point.index = count  # PKHG_INFO source renumbered too!

        count += 1

    for facey in source.faces:
        row = len(facey.vertices)
        if row >= 5:
            tmp = Vector()
            for el in facey.vertices:
                tmp = tmp + target.verts[el.index].vector
            tmp = tmp / row
            centre = vertex(tmp)
            centre.index = count  # PKHG_??? give it a good index
            count += 1

            target.verts.append(centre)
            for i in range(row):
                if i == row - 1:
                    a = target.verts[facey.vertices[-1].index]
                    b = target.verts[facey.vertices[0].index]
                else:
                    a = target.verts[facey.vertices[i].index]
                    b = target.verts[facey.vertices[i + 1].index]
                target.faces.append([a, b, centre])
        else:
            f = []

            for j in range(len(facey.vertices)):
                a = facey.vertices[j]
                f.append(target.verts[a.index])

            target.faces.append(f)



# File sequencer_slide_strip.py

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


import bpy
from bpy.types import (
        Operator,
        Panel,
        )
from . import functions


class EcoPanel(Panel):
    bl_label = "Eco Tool"
    bl_idname = "OBJECT_PT_EcoTool"
    bl_space_type = "SEQUENCE_EDITOR"
    bl_region_type = "UI"

    @staticmethod
    def has_sequencer(context):
        return (context.space_data.view_type in
               {'SEQUENCER', 'SEQUENCER_PREVIEW'})

    @classmethod
    def poll(self, context):
        if context.space_data.view_type in {'SEQUENCER', 'SEQUENCER_PREVIEW'}:
            strip = functions.act_strip(context)
            scn = context.scene
            preferences = context.user_preferences
            prefs = preferences.addons[__package__].preferences
            if scn and scn.sequence_editor and scn.sequence_editor.active_strip:
                if prefs.use_eco_tools:
                    return strip.type in ('META')
        else:
            return False

    def draw_header(self, context):
        layout = self.layout
        layout.label(text="", icon="FORCE_HARMONIC")

    def draw(self, context):
        strip = functions.act_strip(context)
        seq_type = strip.type

        preferences = context.user_preferences
        prefs = preferences.addons[__package__].preferences

        if seq_type in ('MOVIE', 'IMAGE', 'META', 'MOVIECLIP', 'SCENE'):
            layout = self.layout
            col = layout.column()

            col.prop(prefs, "eco_value", text="Ecos")
            col.prop(prefs, "eco_offset", text="Offset")
            col.prop(prefs, "eco_use_add_blend_mode", text="Use add blend mode")
            col.operator("sequencer.eco")


class OBJECT_OT_EcoOperator(Operator):
    bl_idname = "sequencer.eco"
    bl_label = "Eco operator"
    bl_description = "Generate an echo effect by duplicating the selected strip"
    bl_options = {'REGISTER', 'UNDO'}

    @staticmethod
    def has_sequencer(context):
        return (context.space_data.view_type in
               {'SEQUENCER', 'SEQUENCER_PREVIEW'})

    @classmethod
    def poll(self, context):
        strip = functions.act_strip(context)
        scn = context.scene
        if scn and scn.sequence_editor and scn.sequence_editor.active_strip:
            return strip.type in ('META')
        else:
            return False

    def execute(self, context):
        active_strip = functions.act_strip(context)

        preferences = context.user_preferences
        prefs = preferences.addons[__package__].preferences

        eco = prefs.eco_value
        offset = prefs.eco_offset

        active_strip.blend_type = 'REPLACE'
        active_strip.blend_alpha = 1
        for i in range(eco):
            bpy.ops.sequencer.duplicate(mode='TRANSLATION')
            bpy.ops.transform.seq_slide(
                    value=(offset, 1), snap=False, snap_target='CLOSEST',
                    snap_point=(0, 0, 0), snap_align=False,
                    snap_normal=(0, 0, 0), release_confirm=False
                    )

            active_strip = functions.act_strip(context)

            if prefs.eco_use_add_blend_mode:
                active_strip.blend_type = 'ADD'
                active_strip.blend_alpha = 1 - 1 / eco
            else:
                active_strip.blend_type = 'ALPHA_OVER'
                active_strip.blend_alpha = 1 / eco

        bpy.ops.sequencer.select_all(action='TOGGLE')
        bpy.ops.sequencer.select_all(action='TOGGLE')
        bpy.ops.sequencer.meta_make()

        return {'FINISHED'}



# <pep8-80 compliant>

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "kgeogeo, mem, Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "4.4"
__date__ = "2 Aug 2017"

import bpy
import bmesh
from mathutils import Vector


class MUV_MVUV(bpy.types.Operator):
    """
    Operator class: Move UV from View3D
    """

    bl_idname = "view3d.muv_mvuv"
    bl_label = "Move the UV from View3D"
    bl_options = {'REGISTER', 'UNDO'}

    def __init__(self):
        self.__topology_dict = []
        self.__prev_mouse = Vector((0.0, 0.0))
        self.__offset_uv = Vector((0.0, 0.0))
        self.__prev_offset_uv = Vector((0.0, 0.0))
        self.__first_time = True
        self.__ini_uvs = []
        self.__running = False

    def __find_uv(self, context):
        bm = bmesh.from_edit_mesh(context.object.data)
        topology_dict = []
        uvs = []
        active_uv = bm.loops.layers.uv.active
        for fidx, f in enumerate(bm.faces):
            for vidx, v in enumerate(f.verts):
                if v.select:
                    uvs.append(f.loops[vidx][active_uv].uv.copy())
                    topology_dict.append([fidx, vidx])

        return topology_dict, uvs

    @classmethod
    def poll(cls, context):
        return context.edit_object

    def modal(self, context, event):
        if self.__first_time is True:
            self.__prev_mouse = Vector((
                event.mouse_region_x, event.mouse_region_y))
            self.__first_time = False
            return {'RUNNING_MODAL'}

        # move UV
        div = 10000
        self.__offset_uv += Vector((
            (event.mouse_region_x - self.__prev_mouse.x) / div,
            (event.mouse_region_y - self.__prev_mouse.y) / div))
        ouv = self.__offset_uv
        pouv = self.__prev_offset_uv
        vec = Vector((ouv.x - ouv.y, ouv.x + ouv.y))
        dv = vec - pouv
        self.__prev_offset_uv = vec
        self.__prev_mouse = Vector((
            event.mouse_region_x, event.mouse_region_y))

        # check if operation is started
        if self.__running is True:
            if event.type == 'LEFTMOUSE' and event.value == 'RELEASE':
                self.__running = False
            return {'RUNNING_MODAL'}

        # update UV
        obj = context.object
        bm = bmesh.from_edit_mesh(obj.data)
        active_uv = bm.loops.layers.uv.active
        for fidx, vidx in self.__topology_dict:
            l = bm.faces[fidx].loops[vidx]
            l[active_uv].uv = l[active_uv].uv + dv
        bmesh.update_edit_mesh(obj.data)

        # check mouse preference
        if context.user_preferences.inputs.select_mouse == 'RIGHT':
            confirm_btn = 'LEFTMOUSE'
            cancel_btn = 'RIGHTMOUSE'
        else:
            confirm_btn = 'RIGHTMOUSE'
            cancel_btn = 'LEFTMOUSE'

        # cancelled
        if event.type == cancel_btn and event.value == 'PRESS':
            for (fidx, vidx), uv in zip(self.__topology_dict, self.__ini_uvs):
                bm.faces[fidx].loops[vidx][active_uv].uv = uv
            return {'FINISHED'}
        # confirmed
        if event.type == confirm_btn and event.value == 'PRESS':
            return {'FINISHED'}

        return {'RUNNING_MODAL'}

    def execute(self, context):
        self.__first_time = True
        self.__running = True
        context.window_manager.modal_handler_add(self)
        self.__topology_dict, self.__ini_uvs = self.__find_uv(context)
        return {'RUNNING_MODAL'}



# <pep8-80 compliant>

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "4.4"
__date__ = "2 Aug 2017"

from collections import namedtuple

import bpy
import bgl
import bmesh
import mathutils
from bpy_extras import view3d_utils

from . import muv_common


Rect = namedtuple('Rect', 'x0 y0 x1 y1')
Rect2 = namedtuple('Rect2', 'x y width height')


def get_canvas(context, magnitude):
    """
    Get canvas to be renderred texture
    """
    sc = context.scene
    prefs = context.user_preferences.addons["uv_magic_uv"].preferences

    region_w = context.region.width
    region_h = context.region.height
    canvas_w = region_w - prefs.texproj_canvas_padding[0] * 2.0
    canvas_h = region_h - prefs.texproj_canvas_padding[1] * 2.0

    img = bpy.data.images[sc.muv_texproj_tex_image]
    tex_w = img.size[0]
    tex_h = img.size[1]

    center_x = region_w * 0.5
    center_y = region_h * 0.5

    if sc.muv_texproj_adjust_window:
        ratio_x = canvas_w / tex_w
        ratio_y = canvas_h / tex_h
        if sc.muv_texproj_apply_tex_aspect:
            ratio = ratio_y if ratio_x > ratio_y else ratio_x
            len_x = ratio * tex_w
            len_y = ratio * tex_h
        else:
            len_x = canvas_w
            len_y = canvas_h
    else:
        if sc.muv_texproj_apply_tex_aspect:
            len_x = tex_w * magnitude
            len_y = tex_h * magnitude
        else:
            len_x = region_w * magnitude
            len_y = region_h * magnitude

    x0 = int(center_x - len_x * 0.5)
    y0 = int(center_y - len_y * 0.5)
    x1 = int(center_x + len_x * 0.5)
    y1 = int(center_y + len_y * 0.5)

    return Rect(x0, y0, x1, y1)


def rect_to_rect2(rect):
    """
    Convert Rect1 to Rect2
    """

    return Rect2(rect.x0, rect.y0, rect.x1 - rect.x0, rect.y1 - rect.y0)


def region_to_canvas(rg_vec, canvas):
    """
    Convert screen region to canvas
    """

    cv_rect = rect_to_rect2(canvas)
    cv_vec = mathutils.Vector()
    cv_vec.x = (rg_vec.x - cv_rect.x) / cv_rect.width
    cv_vec.y = (rg_vec.y - cv_rect.y) / cv_rect.height

    return cv_vec


class MUV_TexProjRenderer(bpy.types.Operator):
    """
    Operation class: Render selected texture
    No operation (only rendering texture)
    """

    bl_idname = "uv.muv_texproj_renderer"
    bl_description = "Render selected texture"
    bl_label = "Texture renderer"

    __handle = None

    @staticmethod
    def handle_add(obj, context):
        MUV_TexProjRenderer.__handle = bpy.types.SpaceView3D.draw_handler_add(
            MUV_TexProjRenderer.draw_texture,
            (obj, context), 'WINDOW', 'POST_PIXEL')

    @staticmethod
    def handle_remove():
        if MUV_TexProjRenderer.__handle is not None:
            bpy.types.SpaceView3D.draw_handler_remove(
                MUV_TexProjRenderer.__handle, 'WINDOW')
            MUV_TexProjRenderer.__handle = None

    @staticmethod
    def draw_texture(_, context):
        sc = context.scene

        # no textures are selected
        if sc.muv_texproj_tex_image == "None":
            return

        # get texture to be renderred
        img = bpy.data.images[sc.muv_texproj_tex_image]

        # setup rendering region
        rect = get_canvas(context, sc.muv_texproj_tex_magnitude)
        positions = [
            [rect.x0, rect.y0],
            [rect.x0, rect.y1],
            [rect.x1, rect.y1],
            [rect.x1, rect.y0]
        ]
        tex_coords = [
            [0.0, 0.0],
            [0.0, 1.0],
            [1.0, 1.0],
            [1.0, 0.0]
        ]

        # OpenGL configuration
        bgl.glEnable(bgl.GL_BLEND)
        bgl.glEnable(bgl.GL_TEXTURE_2D)
        if img.bindcode:
            bind = img.bindcode[0]
            bgl.glBindTexture(bgl.GL_TEXTURE_2D, bind)
            bgl.glTexParameteri(
                bgl.GL_TEXTURE_2D, bgl.GL_TEXTURE_MIN_FILTER, bgl.GL_LINEAR)
            bgl.glTexParameteri(
                bgl.GL_TEXTURE_2D, bgl.GL_TEXTURE_MAG_FILTER, bgl.GL_LINEAR)
            bgl.glTexEnvi(
                bgl.GL_TEXTURE_ENV, bgl.GL_TEXTURE_ENV_MODE, bgl.GL_MODULATE)

        # render texture
        bgl.glBegin(bgl.GL_QUADS)
        bgl.glColor4f(1.0, 1.0, 1.0, sc.muv_texproj_tex_transparency)
        for (v1, v2), (u, v) in zip(positions, tex_coords):
            bgl.glTexCoord2f(u, v)
            bgl.glVertex2f(v1, v2)
        bgl.glEnd()


class MUV_TexProjStart(bpy.types.Operator):
    """
    Operation class: Start Texture Projection
    """

    bl_idname = "uv.muv_texproj_start"
    bl_label = "Start Texture Projection"
    bl_description = "Start Texture Projection"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        props = context.scene.muv_props.texproj
        if props.running is False:
            MUV_TexProjRenderer.handle_add(self, context)
            props.running = True
        if context.area:
            context.area.tag_redraw()

        return {'FINISHED'}


class MUV_TexProjStop(bpy.types.Operator):
    """
    Operation class: Stop Texture Projection
    """

    bl_idname = "uv.muv_texproj_stop"
    bl_label = "Stop Texture Projection"
    bl_description = "Stop Texture Projection"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        props = context.scene.muv_props.texproj
        if props.running is True:
            MUV_TexProjRenderer.handle_remove()
            props.running = False
        if context.area:
            context.area.tag_redraw()

        return {'FINISHED'}


class MUV_TexProjProject(bpy.types.Operator):
    """
    Operation class: Project texture
    """

    bl_idname = "uv.muv_texproj_project"
    bl_label = "Project Texture"
    bl_description = "Project Texture"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj is not None and obj.type == "MESH"

    def execute(self, context):
        sc = context.scene

        if context.mode != "EDIT_MESH":
            self.report({'WARNING'}, "Mesh must be in Edit mode")
            return {'CANCELLED'}

        if sc.muv_texproj_tex_image == "None":
            self.report({'WARNING'}, "No textures are selected")
            return {'CANCELLED'}

        _, region, space = muv_common.get_space(
            'VIEW_3D', 'WINDOW', 'VIEW_3D')

        # get faces to be texture projected
        obj = context.active_object
        world_mat = obj.matrix_world
        bm = bmesh.from_edit_mesh(obj.data)
        if muv_common.check_version(2, 73, 0) >= 0:
            bm.faces.ensure_lookup_table()

        # get UV and texture layer
        if not bm.loops.layers.uv:
            self.report({'WARNING'}, "Object must have more than one UV map")
            return {'CANCELLED'}

        uv_layer = bm.loops.layers.uv.verify()
        tex_layer = bm.faces.layers.tex.verify()

        sel_faces = [f for f in bm.faces if f.select]

        # transform 3d space to screen region
        v_screen = [
            view3d_utils.location_3d_to_region_2d(
                region,
                space.region_3d,
                world_mat * l.vert.co)
            for f in sel_faces for l in f.loops
        ]

        # transform screen region to canvas
        v_canvas = [
            region_to_canvas(
                v,
                get_canvas(bpy.context, sc.muv_texproj_tex_magnitude))
            for v in v_screen
        ]

        # project texture to object
        i = 0
        for f in sel_faces:
            f[tex_layer].image = bpy.data.images[sc.muv_texproj_tex_image]
            for l in f.loops:
                l[uv_layer].uv = v_canvas[i].to_2d()
                i = i + 1

        muv_common.redraw_all_areas()
        bmesh.update_edit_mesh(obj.data)

        return {'FINISHED'}


class OBJECT_PT_TP(bpy.types.Panel):
    """
    Panel class: Texture Projection Menu on Property Panel on View3D
    """

    bl_label = "Texture Projection"
    bl_description = "Texture Projection Menu"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_context = 'mesh_edit'

    @classmethod
    def poll(cls, context):
        prefs = context.user_preferences.addons["uv_magic_uv"].preferences
        return prefs.enable_texproj

    def draw_header(self, _):
        layout = self.layout
        layout.label(text="", icon='IMAGE_COL')

    def draw(self, context):
        sc = context.scene
        layout = self.layout
        props = sc.muv_props.texproj
        if props.running is False:
            layout.operator(
                MUV_TexProjStart.bl_idname, text="Start", icon='PLAY')
        else:
            layout.operator(
                MUV_TexProjStop.bl_idname, text="Stop", icon='PAUSE')
            layout.prop(sc, "muv_texproj_tex_image", text="Image")
            layout.prop(
                sc, "muv_texproj_tex_transparency", text="Transparency"
            )
            layout.prop(sc, "muv_texproj_adjust_window", text="Adjust Window")
            if not sc.muv_texproj_adjust_window:
                layout.prop(sc, "muv_texproj_tex_magnitude", text="Magnitude")
            layout.prop(
                sc, "muv_texproj_apply_tex_aspect", text="Texture Aspect Ratio"
            )
            layout.operator(MUV_TexProjProject.bl_idname, text="Project")



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8-80 compliant>

__all__ = (
    "mesh_linked_uv_islands",
    "mesh_linked_tessfaces",
    "edge_face_count_dict",
    "edge_face_count",
    "edge_loops_from_tessfaces",
    "edge_loops_from_edges",
    "ngon_tessellate",
    "face_random_points",
    )


def mesh_linked_uv_islands(mesh):
    """
    Splits the mesh into connected polygons, use this for seperating cubes from
    other mesh elements within 1 mesh datablock.

    :arg mesh: the mesh used to group with.
    :type mesh: :class:`bpy.types.Mesh`
    :return: lists of lists containing polygon indices
    :rtype: list
    """
    uv_loops = [luv.uv[:] for luv in mesh.uv_layers.active.data]
    poly_loops = [poly.loop_indices for poly in mesh.polygons]
    luv_hash = {}
    luv_hash_get = luv_hash.get
    luv_hash_ls = [None] * len(uv_loops)
    for pi, poly_indices in enumerate(poly_loops):
        for li in poly_indices:
            uv = uv_loops[li]
            uv_hub = luv_hash_get(uv)
            if uv_hub is None:
                uv_hub = luv_hash[uv] = [pi]
            else:
                uv_hub.append(pi)
            luv_hash_ls[li] = uv_hub

    poly_islands = []

    # 0 = none, 1 = added, 2 = searched
    poly_tag = [0] * len(poly_loops)

    while True:
        poly_index = -1
        for i in range(len(poly_loops)):
            if poly_tag[i] == 0:
                poly_index = i
                break

        if poly_index != -1:
            island = [poly_index]
            poly_tag[poly_index] = 1
            poly_islands.append(island)
        else:
            break  # we're done

        added = True
        while added:
            added = False
            for poly_index in island[:]:
                if poly_tag[poly_index] == 1:
                    for li in poly_loops[poly_index]:
                        for poly_index_shared in luv_hash_ls[li]:
                            if poly_tag[poly_index_shared] == 0:
                                added = True
                                poly_tag[poly_index_shared] = 1
                                island.append(poly_index_shared)
                    poly_tag[poly_index] = 2

    return poly_islands


def mesh_linked_tessfaces(mesh):
    """
    Splits the mesh into connected faces, use this for seperating cubes from
    other mesh elements within 1 mesh datablock.

    :arg mesh: the mesh used to group with.
    :type mesh: :class:`bpy.types.Mesh`
    :return: lists of lists containing faces.
    :rtype: list
    """

    # Build vert face connectivity
    vert_faces = [[] for i in range(len(mesh.vertices))]
    for f in mesh.tessfaces:
        for v in f.vertices:
            vert_faces[v].append(f)

    # sort faces into connectivity groups
    face_groups = [[f] for f in mesh.tessfaces]
    # map old, new face location
    face_mapping = list(range(len(mesh.tessfaces)))

    # Now clump faces iteratively
    ok = True
    while ok:
        ok = False

        for i, f in enumerate(mesh.tessfaces):
            mapped_index = face_mapping[f.index]
            mapped_group = face_groups[mapped_index]

            for v in f.vertices:
                for nxt_f in vert_faces[v]:
                    if nxt_f != f:
                        nxt_mapped_index = face_mapping[nxt_f.index]

                        # We are not a part of the same group
                        if mapped_index != nxt_mapped_index:
                            ok = True

                            # Assign mapping to this group so they
                            # all map to this group
                            for grp_f in face_groups[nxt_mapped_index]:
                                face_mapping[grp_f.index] = mapped_index

                            # Move faces into this group
                            mapped_group.extend(face_groups[nxt_mapped_index])

                            # remove reference to the list
                            face_groups[nxt_mapped_index] = None

    # return all face groups that are not null
    # this is all the faces that are connected in their own lists.
    return [fg for fg in face_groups if fg]


def edge_face_count_dict(mesh):
    """
    :return: dict of edge keys with their value set to the number of
       faces using each edge.
    :rtype: dict
    """

    face_edge_count = {}
    loops = mesh.loops
    edges = mesh.edges
    for poly in mesh.polygons:
        for i in poly.loop_indices:
            key = edges[loops[i].edge_index].key
            try:
                face_edge_count[key] += 1
            except:
                face_edge_count[key] = 1

    return face_edge_count


def edge_face_count(mesh):
    """
    :return: list face users for each item in mesh.edges.
    :rtype: list
    """
    edge_face_count = edge_face_count_dict(mesh)
    get = dict.get
    return [get(edge_face_count, ed.key, 0) for ed in mesh.edges]


def edge_loops_from_tessfaces(mesh, tessfaces=None, seams=()):
    """
    Edge loops defined by faces

    Takes me.tessfaces or a list of faces and returns the edge loops
    These edge loops are the edges that sit between quads, so they dont touch
    1 quad, note: not connected will make 2 edge loops,
    both only containing 2 edges.

    return a list of edge key lists
    [[(0, 1), (4, 8), (3, 8)], ...]

    :arg mesh: the mesh used to get edge loops from.
    :type mesh: :class:`bpy.types.Mesh`
    :arg tessfaces: optional face list to only use some of the meshes faces.
    :type tessfaces: :class:`bpy.types.MeshTessFace`, sequence or or NoneType
    :return: return a list of edge vertex index lists.
    :rtype: list
    """

    OTHER_INDEX = 2, 3, 0, 1  # opposite face index

    if tessfaces is None:
        tessfaces = mesh.tessfaces

    edges = {}

    for f in tessfaces:
        if len(f.vertices) == 4:
            edge_keys = f.edge_keys
            for i, edkey in enumerate(f.edge_keys):
                edges.setdefault(edkey, []).append(edge_keys[OTHER_INDEX[i]])

    for edkey in seams:
        edges[edkey] = []

    # Collect edge loops here
    edge_loops = []

    for edkey, ed_adj in edges.items():
        if 0 < len(ed_adj) < 3:  # 1 or 2
            # Seek the first edge
            context_loop = [edkey, ed_adj[0]]
            edge_loops.append(context_loop)
            if len(ed_adj) == 2:
                other_dir = ed_adj[1]
            else:
                other_dir = None

            del ed_adj[:]

            flipped = False

            while 1:
                # from knowing the last 2, look for the next.
                ed_adj = edges[context_loop[-1]]
                if len(ed_adj) != 2:
                    # the original edge had 2 other edges
                    if other_dir and flipped is False:
                        flipped = True  # only flip the list once
                        context_loop.reverse()
                        del ed_adj[:]
                        context_loop.append(other_dir)  # save 1 look-up

                        ed_adj = edges[context_loop[-1]]
                        if len(ed_adj) != 2:
                            del ed_adj[:]
                            break
                    else:
                        del ed_adj[:]
                        break

                i = ed_adj.index(context_loop[-2])
                context_loop.append(ed_adj[not i])

                # Dont look at this again
                del ed_adj[:]

    return edge_loops


def edge_loops_from_edges(mesh, edges=None):
    """
    Edge loops defined by edges

    Takes me.edges or a list of edges and returns the edge loops

    return a list of vertex indices.
    [ [1, 6, 7, 2], ...]

    closed loops have matching start and end values.
    """
    line_polys = []

    # Get edges not used by a face
    if edges is None:
        edges = mesh.edges

    if not hasattr(edges, "pop"):
        edges = edges[:]

    while edges:
        current_edge = edges.pop()
        vert_end, vert_start = current_edge.vertices[:]
        line_poly = [vert_start, vert_end]

        ok = True
        while ok:
            ok = False
            #for i, ed in enumerate(edges):
            i = len(edges)
            while i:
                i -= 1
                ed = edges[i]
                v1, v2 = ed.vertices
                if v1 == vert_end:
                    line_poly.append(v2)
                    vert_end = line_poly[-1]
                    ok = 1
                    del edges[i]
                    # break
                elif v2 == vert_end:
                    line_poly.append(v1)
                    vert_end = line_poly[-1]
                    ok = 1
                    del edges[i]
                    #break
                elif v1 == vert_start:
                    line_poly.insert(0, v2)
                    vert_start = line_poly[0]
                    ok = 1
                    del edges[i]
                    # break
                elif v2 == vert_start:
                    line_poly.insert(0, v1)
                    vert_start = line_poly[0]
                    ok = 1
                    del edges[i]
                    #break
        line_polys.append(line_poly)

    return line_polys


def ngon_tessellate(from_data, indices, fix_loops=True):
    """
    Takes a polyline of indices (fgon) and returns a list of face
    index lists. Designed to be used for importers that need indices for an
    fgon to create from existing verts.

    :arg from_data: either a mesh, or a list/tuple of vectors.
    :type from_data: list or :class:`bpy.types.Mesh`
    :arg indices: a list of indices to use this list
       is the ordered closed polyline
       to fill, and can be a subset of the data given.
    :type indices: list
    :arg fix_loops: If this is enabled polylines
       that use loops to make multiple
       polylines are delt with correctly.
    :type fix_loops: bool
    """

    from mathutils.geometry import tessellate_polygon
    from mathutils import Vector
    vector_to_tuple = Vector.to_tuple

    if not indices:
        return []

    def mlen(co):
        # manhatten length of a vector, faster then length
        return abs(co[0]) + abs(co[1]) + abs(co[2])

    def vert_treplet(v, i):
        return v, vector_to_tuple(v, 6), i, mlen(v)

    def ed_key_mlen(v1, v2):
        if v1[3] > v2[3]:
            return v2[1], v1[1]
        else:
            return v1[1], v2[1]

    if not fix_loops:
        """
        Normal single concave loop filling
        """
        if type(from_data) in {tuple, list}:
            verts = [Vector(from_data[i]) for ii, i in enumerate(indices)]
        else:
            verts = [from_data.vertices[i].co for ii, i in enumerate(indices)]

        # same as reversed(range(1, len(verts))):
        for i in range(len(verts) - 1, 0, -1):
            if verts[i][1] == verts[i - 1][0]:
                verts.pop(i - 1)

        fill = tessellate_polygon([verts])

    else:
        """
        Seperate this loop into multiple loops be finding edges that are
        used twice. This is used by lightwave LWO files a lot
        """

        if type(from_data) in {tuple, list}:
            verts = [vert_treplet(Vector(from_data[i]), ii)
                     for ii, i in enumerate(indices)]
        else:
            verts = [vert_treplet(from_data.vertices[i].co, ii)
                     for ii, i in enumerate(indices)]

        edges = [(i, i - 1) for i in range(len(verts))]
        if edges:
            edges[0] = (0, len(verts) - 1)

        if not verts:
            return []

        edges_used = set()
        edges_doubles = set()
        # We need to check if any edges are used twice location based.
        for ed in edges:
            edkey = ed_key_mlen(verts[ed[0]], verts[ed[1]])
            if edkey in edges_used:
                edges_doubles.add(edkey)
            else:
                edges_used.add(edkey)

        # Store a list of unconnected loop segments split by double edges.
        # will join later
        loop_segments = []

        v_prev = verts[0]
        context_loop = [v_prev]
        loop_segments = [context_loop]

        for v in verts:
            if v != v_prev:
                # Are we crossing an edge we removed?
                if ed_key_mlen(v, v_prev) in edges_doubles:
                    context_loop = [v]
                    loop_segments.append(context_loop)
                else:
                    if context_loop and context_loop[-1][1] == v[1]:
                        pass
                    else:
                        context_loop.append(v)

                v_prev = v
        # Now join loop segments

        def join_seg(s1, s2):
            if s2[-1][1] == s1[0][1]:
                s1, s2 = s2, s1
            elif s1[-1][1] == s2[0][1]:
                pass
            else:
                return False

            # If were stuill here s1 and s2 are 2 segments in the same polyline
            s1.pop()  # remove the last vert from s1
            s1.extend(s2)  # add segment 2 to segment 1

            if s1[0][1] == s1[-1][1]:  # remove endpoints double
                s1.pop()

            del s2[:]  # Empty this segment s2 so we don't use it again.
            return True

        joining_segments = True
        while joining_segments:
            joining_segments = False
            segcount = len(loop_segments)

            for j in range(segcount - 1, -1, -1):  # reversed(range(segcount)):
                seg_j = loop_segments[j]
                if seg_j:
                    for k in range(j - 1, -1, -1):  # reversed(range(j)):
                        if not seg_j:
                            break
                        seg_k = loop_segments[k]

                        if seg_k and join_seg(seg_j, seg_k):
                            joining_segments = True

        loop_list = loop_segments

        for verts in loop_list:
            while verts and verts[0][1] == verts[-1][1]:
                verts.pop()

        loop_list = [verts for verts in loop_list if len(verts) > 2]
        # DONE DEALING WITH LOOP FIXING

        # vert mapping
        vert_map = [None] * len(indices)
        ii = 0
        for verts in loop_list:
            if len(verts) > 2:
                for i, vert in enumerate(verts):
                    vert_map[i + ii] = vert[2]
                ii += len(verts)

        fill = tessellate_polygon([[v[0] for v in loop] for loop in loop_list])
        #draw_loops(loop_list)
        #raise Exception("done loop")
        # map to original indices
        fill = [[vert_map[i] for i in reversed(f)] for f in fill]

    if not fill:
        print('Warning Cannot scanfill, fallback on a triangle fan.')
        fill = [[0, i - 1, i] for i in range(2, len(indices))]
    else:
        # Use real scanfill.
        # See if its flipped the wrong way.
        flip = None
        for fi in fill:
            if flip is not None:
                break
            for i, vi in enumerate(fi):
                if vi == 0 and fi[i - 1] == 1:
                    flip = False
                    break
                elif vi == 1 and fi[i - 1] == 0:
                    flip = True
                    break

        if not flip:
            for i, fi in enumerate(fill):
                fill[i] = tuple([ii for ii in reversed(fi)])

    return fill


def face_random_points(num_points, tessfaces):
    """
    Generates a list of random points over mesh tessfaces.

    :arg num_points: the number of random points to generate on each face.
    :type int:
    :arg tessfaces: list of the faces to generate points on.
    :type tessfaces: :class:`bpy.types.MeshTessFace`, sequence
    :return: list of random points over all faces.
    :rtype: list
    """

    from random import random
    from mathutils.geometry import area_tri

    # Split all quads into 2 tris, tris remain unchanged
    tri_faces = []
    for f in tessfaces:
        tris = []
        verts = f.id_data.vertices
        fv = f.vertices[:]
        tris.append((verts[fv[0]].co,
                     verts[fv[1]].co,
                     verts[fv[2]].co,
                     ))
        if len(fv) == 4:
            tris.append((verts[fv[0]].co,
                         verts[fv[3]].co,
                         verts[fv[2]].co,
                         ))
        tri_faces.append(tris)

    # For each face, generate the required number of random points
    sampled_points = [None] * (num_points * len(tessfaces))
    for i, tf in enumerate(tri_faces):
        for k in range(num_points):
            # If this is a quad, we need to weight its 2 tris by their area
            if len(tf) != 1:
                area1 = area_tri(*tf[0])
                area2 = area_tri(*tf[1])
                area_tot = area1 + area2

                area1 = area1 / area_tot
                area2 = area2 / area_tot

                vecs = tf[0 if (random() < area1) else 1]
            else:
                vecs = tf[0]

            u1 = random()
            u2 = random()
            u_tot = u1 + u2

            if u_tot > 1:
                u1 = 1.0 - u1
                u2 = 1.0 - u2

            side1 = vecs[1] - vecs[0]
            side2 = vecs[2] - vecs[0]

            p = vecs[0] + u1 * side1 + u2 * side2

            sampled_points[num_points * i + k] = p

    return sampled_points



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
import os.path
import operator
import subprocess
import random
from bpy.props import (
        IntProperty,
        FloatProperty,
        EnumProperty,
        BoolProperty,
        )

imb_ext_image = [
    # IMG
    ".png", ".tga", ".bmp", ".jpg", ".jpeg", ".sgi", ".rgb",
    ".rgba", ".tif", ".tiff", ".tx", ".jp2", ".hdr", ".dds",
    ".dpx", ".cin", ".exr", ".rw2",
    # IMG QT
    ".gif", ".psd", ".pct", ".pict", ".pntg", ".qtif"
    ]
imb_ext_audio = [
    ".wav", ".ogg", ".oga", ".mp3", ".mp2", ".ac3", ".aac",
    ".flac", ".wma", ".eac3", ".aif", ".aiff", ".m4a"
    ]
imb_ext_movie = [
    ".avi", ".flc", ".mov", ".movie", ".mp4", ".m4v", ".m2v",
    ".m2t", ".m2ts", ".mts", ".mv", ".avs", ".wmv", ".ogv", ".ogg",
    ".dv", ".mpeg", ".mpg", ".mpg2", ".vob", ".mkv", ".flv",
    ".divx", ".xvid", ".mxf"
    ]
movieextdict = [
    ("1", ".avi", ""),
    ("2", ".flc", ""), ("3", ".mov", ""),
    ("4", ".movie", ""), ("5", ".mp4", ""),
    ("6", ".m4v", ""), ("7", ".m2v", ""),
    ("8", ".m2t", ""), ("9", ".m2ts", ""),
    ("10", ".mts", ""), ("11", ".mv", ""),
    ("12", ".avs", ""), ("13", ".wmv", ""),
    ("14", ".ogv", ""), ("15", ".dv", ""),
    ("16", ".mpeg", ""), ("17", ".mpg", ""),
    ("18", ".mpg2", ""), ("19", ".vob", ""),
    ("20", ".mkv", ""), ("21", ".flv", ""),
    ("22", ".divx", ""), ("23", ".xvid", ""),
    ("24", ".mxf", "")
    ]


# Functions
def error_handlers(self, op_name, errors, reports="ERROR"):
    if self and reports:
        self.report({'INFO'},
                    reports + ": some operations could not be performed "
                    "(See Console for more info)")

    print("\n[Kinoraw Tools]\nOperator: {}\nWarning: {}\n".format(op_name, errors))


def initSceneProperties(context):
    # initSceneProperties is ONLY for varaibles that should
    # be keeped with the blend file. Any other addon preferences
    # should go to the addon preferences operator in __init__
    try:
        if context.scene.kr_scn_init is True:
            return False
    except AttributeError:
        pass

    scn = context.scene

    # jump to cut
    bpy.types.Scene.kr_auto_markers = BoolProperty(
            name="kr_auto_markers",
            description="Activate Auto markers",
            default=False
            )
    scn.kr_auto_markers = False

    bpy.types.Scene.kr_in_marker = IntProperty(
            name="In",
            description="In frame position",
            min=-30000, max=30000,
            default=1
            )
    scn.kr_in_marker = 1

    bpy.types.Scene.kr_out_marker = IntProperty(
            name="Out",
            description="Out frame position",
            min=scn.kr_in_marker, max=30000,
            default=75
            )
    scn.kr_out_marker = 75

    # sequencer extra actions
    bpy.types.Scene.kr_default_fade_duration = IntProperty(
            name="Duration",
            description="Number of frames to fade",
            min=1, max=250,
            default=scn.render.fps
            )
    scn.kr_default_fade_duration = scn.render.fps

    bpy.types.Scene.kr_default_fade_amount = FloatProperty(
            name="Amount",
            description="Maximum value of fade",
            min=0.0,
            max=100.0,
            default=1.0
            )
    scn.kr_default_fade_amount = 1.0

    # recursive loader
    bpy.types.Scene.kr_recursive = BoolProperty(
            name="Recursive",
            description="Load in recursive folders",
            default=False
            )
    scn.kr_recursive = False

    bpy.types.Scene.kr_recursive_select_by_extension = BoolProperty(
            name="Recursive ext",
            description="Load only clips with selected extension",
            default=False
            )
    scn.kr_recursive_select_by_extension = False

    bpy.types.Scene.kr_default_ext = EnumProperty(
            items=movieextdict,
            name="ext enum",
            default="3"
            )
    scn.kr_default_ext = "3"

    bpy.types.Scene.kr_scn_init = BoolProperty(
            name="Init",
            default=False
            )
    scn.kr_scn_init = True

    return True


def get_selected_strips(context):
    "return a list of selected strips"
    strips = []
    for i in context.scene.sequence_editor.sequences_all:
        if i.select is True:
            strips.append(i)
    return strips


def create_folder(path):
    if not os.path.isdir(bpy.path.abspath(path)):
        folder = bpy.path.abspath(path)
        command = "mkdir " + folder
        subprocess.call(command, shell=True)


def add_marker(context, text, frame):
    scn = context.scene
    markers = scn.timeline_markers
    mark = markers.new(name=text)
    mark.frame = frame


def act_strip(context):
    try:
        return context.scene.sequence_editor.active_strip
    except AttributeError:
        return None


def detect_strip_type(filepath):
    extension = os.path.splitext(filepath)[1]
    extension = extension.lower()
    if extension in imb_ext_image:
        type = 'IMAGE'
    elif extension in imb_ext_movie:
        type = 'MOVIE'
    elif extension in imb_ext_audio:
        type = 'SOUND'
    else:
        type = None

    return type


# recursive load functions
def getpathfrombrowser(context):
    '''
    returns path from filebrowser
    '''
    for a in context.window.screen.areas:
        if a.type == 'FILE_BROWSER':
            params = a.spaces[0].params
            break
    try:
        params
    except UnboundLocalError:
        return {'CANCELLED'}

    path = params.directory
    return path


def getfilepathfrombrowser(context):
    """
    returns path and file from filebrowser
    """
    for a in context.window.screen.areas:
        if a.type == 'FILE_BROWSER':
            params = a.spaces[0].params
            break
    try:
        params
    except UnboundLocalError:
        return {'CANCELLED'}

    if params.filename == '':
        return {'CANCELLED'}

    path = params.directory
    filename = params.filename
    return path, filename


def setpathinbrowser(context, path, file):
    '''
    set path and file in the filebrowser
    '''
    for a in context.window.screen.areas:
        if a.type == 'FILE_BROWSER':
            params = a.spaces[0].params
            break
    try:
        params
    except UnboundLocalError:

        return {'CANCELLED'}

    params.directory = path
    params.filename = file
    return path, params


def sortlist(filelist):
    '''
    given a list of tuplas (path, filename) returns a list sorted by filename
    '''
    filelist_sorted = sorted(filelist, key=operator.itemgetter(1))
    return filelist_sorted


def onefolder(context, recursive_select_by_extension, ext):
    '''
    returns a list of MOVIE type files from folder selected in file browser
    '''
    filelist = []
    path, filename = getfilepathfrombrowser(context)

    for i in movieextdict:
        if i[0] == ext:
            extension = i[1].rpartition(".")[2]
            break

    if detect_strip_type(path + filename) == 'MOVIE':
        if recursive_select_by_extension is True:
            # filtering by extension...
            for file in os.listdir(path):
                if file.rpartition(".")[2].lower() == extension:
                    filelist.append((path, file))
        else:
            # looking for all known extensions
            for file in os.listdir(path):
                for i in movieextdict:
                    if file.rpartition(".")[2].lower() == i[1].rpartition(".")[2]:
                        filelist.append((path, file))
    return (filelist)


def recursive(context, recursive_select_by_extension, ext):
    '''
    returns a list of MOVIE type files recursively from file browser
    '''
    filelist = []
    path = getpathfrombrowser(context)

    for i in movieextdict:
        if i[0] == ext:
            extension = i[1].rpartition(".")[2]
            break

    for root, dirs, files in os.walk(path):
        for file in files:
            if recursive_select_by_extension is True:
                # filtering by extension...
                if file.rpartition(".")[2].lower() == extension:
                    filelist.append((root, file))
            else:
                # looking for all known extensions
                for i in movieextdict:
                    if file.rpartition(".")[2].lower() == i[1].rpartition(".")[2]:
                        filelist.append((root, file))
    return filelist


# jump to cut functions
def triminout(strip, sin, sout):

    """trim the strip to in and out, and returns
    true if the strip is outside given in and out"""

    start = strip.frame_start + strip.frame_offset_start - strip.frame_still_start
    end = start + strip.frame_final_duration

    remove = False
    if end < sin:
        remove = True
    if start > sout:
        remove = True

    if end > sin:
        if start < sin:
            strip.select_right_handle = False
            strip.select_left_handle = True
            bpy.ops.sequencer.snap(frame=sin)
            strip.select_left_handle = False
    if start < sout:
        if end > sout:
            strip.select_left_handle = False
            strip.select_right_handle = True
            bpy.ops.sequencer.snap(frame=sout)
            strip.select_right_handle = False

    return remove


# random editor functions

def randompartition(lst, n, rand):
    division = len(lst) / float(n)
    lista = []
    for i in range(n):
        lista.append(division)

    var = 0
    for i in range(n - 1):
        lista[i] += random.randint(-int(rand * division), int(rand * division))
        var += lista[i]

    if lista[n - 1] != len(lst) - var:
        lista[n - 1] = len(lst) - var

    random.shuffle(lista)
    division = len(lst) / float(n)
    count = 0
    newlist = []
    for i in range(n):
        # print(lst[count : int(lista[i]-1)+count])
        newlist.append([lst[count: int(lista[i] - 1) + count]])
        count += int(lista[i])

    return newlist


def randomframe(strip):
    # random frame between a and b
    a = strip.frame_start
    b = strip.frame_final_duration
    rand = a + int(random.random() * b)

    return rand


# ???
def get_matching_markers(scene, name=None):
    '''
    return a list of markers with same name
    from the scene, or all markers if name is None
    '''
    selected_markers = []
    markers = scene.timeline_markers
    for mark in markers:
        # print(mark.name, name)
        if mark.name == name or name is None:
            selected_markers.append(mark.frame)

    return selected_markers


def generate_subsets_list(number_of_subsets):
    # generate marker subsets list
    subset_list = []
    subset_names = ['A', 'B', 'C', 'D', 'E', 'F']

    for subset in range(number_of_subsets):
        subset_list.append(subset_names[subset])
    return subset_list


def get_marker_dict(scene, number_of_subsets):
    """
    return a dict where:
        keys = subset names
        values = list of markers
    """

    subset_list = generate_subsets_list(number_of_subsets)
    # generate dict with a list for each subset
    marker_dict = {}

    for subset in subset_list:
        lists = get_matching_markers(scene, subset)
        marker_dict[subset] = lists
    return marker_dict


def get_cut_dict(scene, number_of_subsets):
    """
    return a dict where:
        keys = markers in the scene + start and end
        values = duration in frames from key marker to next marker
    """
    # generate cut_list

    lists = get_matching_markers(scene)
    lists.append(scene.frame_start)
    lists.append(scene.frame_end)
    lists.sort()
    cut_dict = {}

    for i, j in enumerate(lists):
        try:
            cut_dict[j] = lists[i + 1] - j
        except IndexError:
            continue
    return cut_dict



# This file is a part of the HiRISE DTM Importer for Blender
#
# Copyright (C) 2017 Arizona Board of Regents on behalf of the Planetary Image
# Research Laboratory, Lunar and Planetary Laboratory at the University of
# Arizona.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Blender panel for managing a DTM *after* it's been imported"""

import bpy
from bpy.types import (
        Operator,
        Panel,
        )
from bpy.props import FloatProperty

from ..mesh.terrain import BTerrain
from ..mesh.dtm import DTM


class TerrainPanel(Panel):
    """Creates a Panel in the Object properties window for terrain objects"""
    bl_label = "Terrain Model"
    bl_idname = "OBJECT_PT_terrain"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"

    # Allow the user to specify a new resolution factor for reloading the
    # terrain data at. This is useful because it allows the user to stage
    # a scene with a low resolution terrain map, apply textures, modifiers,
    # etc. and then increase the resolution to prepare for final rendering.
    #
    # Displaying this value as a percentage (0, 100] is an intuitive way
    # for users to grasp what this value does. The DTM importer, however,
    # wants to recieve a value between (0, 1]. This is obviously a
    # straightforward conversion:
    #
    #     f(x) = x / 100
    #
    # But this conversion should happen here, in the terrain panel, rather
    # than in the DTM importing utility itself. We can't pass get/set
    # functions to the property itself because they result in a recursion
    # error. Instead, we use another, hidden, property to store the scaled
    # resolution.
    bpy.types.Object.dtm_resolution = FloatProperty(
        subtype="PERCENTAGE",
        name="New Resolution",
        description=(
            "Percentage scale for terrain model resolution. 100\% loads the "
            "model at full resolution (i.e. one vertex for each post in the "
            "original terrain model) and is *MEMORY INTENSIVE*. Downsampling "
            "uses Nearest Neighbors. The downsampling algorithm may need to "
            "alter the resolution you specify here to ensure it results in a "
            "whole number of vertices. If it needs to alter the value you "
            "specify, you are guaranteed that it will shrink it (i.e. "
            "decrease the DTM resolution"
        ),
        min=1.0, max=100.0, default=10.0
    )
    bpy.types.Object.scaled_dtm_resolution = FloatProperty(
        options={'HIDDEN'},
        name="Scaled Terrain Model Resolution",
        get=(lambda self: self.dtm_resolution / 100.0)
    )

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.get("IS_TERRAIN", False)

    def draw(self, context):
        obj = context.active_object
        layout = self.layout

        # User Controls
        layout.prop(obj, 'dtm_resolution')
        layout.operator("terrain.reload")

        # Metadata
        self.draw_metadata_panel(context)

    def draw_metadata_panel(self, context):
        """Display some metadata about the DTM"""
        obj = context.active_object
        layout = self.layout

        metadata_panel = layout.box()

        dtm_resolution = metadata_panel.row()
        dtm_resolution.label('Current Resolution: ')
        dtm_resolution.label('{:9,.2%}'.format(
            obj['DTM_RESOLUTION']
        ))

        mesh_scale = metadata_panel.row()
        mesh_scale.label('Current Scale: ')
        mesh_scale.label('{:9,.2f} m/post'.format(
            obj['MESH_SCALE']
        ))

        dtm_scale = metadata_panel.row()
        dtm_scale.label('Original Scale: ')
        dtm_scale.label('{:9,.2f} m/post'.format(
            obj['MAP_SCALE']
        ))

        return {'FINISHED'}


class ReloadTerrain(Operator):
    """Button for reloading the terrain mesh at a new resolution"""
    bl_idname = "terrain.reload"
    bl_label = "Reload Terrain"

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.get("IS_TERRAIN", False)

    def execute(self, context):
        # Reload the terrain
        obj = context.active_object
        path = obj['PATH']

        scaled_dtm_resolution = obj.scaled_dtm_resolution

        # Reload BTerrain with new DTM
        dtm = DTM(path, scaled_dtm_resolution)
        BTerrain.reload(obj, dtm)

        return {"FINISHED"}



# gpl: author meta-androcto

import bpy
from bpy.types import Operator


class add_texture_scene(Operator):
    bl_idname = "objects_texture.add_scene"
    bl_label = "Create test scene"
    bl_description = "Cycles renderer Scene: Camera aligned to a plane"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        try:
            blend_data = context.blend_data

            # add new scene
            bpy.ops.scene.new(type="NEW")
            scene = bpy.context.scene
            bpy.context.scene.render.engine = 'CYCLES'
            scene.name = "scene_texture_cycles"

            # render settings
            render = scene.render
            render.resolution_x = 1080
            render.resolution_y = 1080
            render.resolution_percentage = 100

            # add new world
            world = bpy.data.worlds.new("Cycles_Textures_World")
            scene.world = world
            world.use_sky_blend = True
            world.use_sky_paper = True
            world.horizon_color = (0.004393, 0.02121, 0.050)
            world.zenith_color = (0.03335, 0.227, 0.359)
            world.light_settings.use_ambient_occlusion = True
            world.light_settings.ao_factor = 0.5

            # add camera
            bpy.ops.view3d.viewnumpad(type='TOP')
            bpy.ops.object.camera_add(
                    location=(0, 0, 2.1850), rotation=(0, 0, 0), view_align=True
                    )
            cam = bpy.context.active_object.data
            cam.lens = 35
            cam.draw_size = 0.1

            # add plane
            bpy.ops.mesh.primitive_plane_add(enter_editmode=True, location=(0, 0, 0))
            bpy.ops.mesh.subdivide(number_cuts=10, smoothness=0)
            bpy.ops.uv.unwrap(method='CONFORMAL', margin=0.001)
            bpy.ops.object.editmode_toggle()
            plane = bpy.context.active_object

            # add plane material
            planeMaterial = blend_data.materials.new("Cycles_Plane_Material")
            bpy.ops.object.material_slot_add()
            plane.material_slots[0].material = planeMaterial
            # Diffuse
            planeMaterial.preview_render_type = "FLAT"
            planeMaterial.diffuse_color = (0.2, 0.2, 0.2)
            # Cycles
            planeMaterial.use_nodes = True

            # Back to Scene
            sc = bpy.context.scene
            bpy.ops.view3d.viewnumpad(type='CAMERA')

        except Exception as e:
            self.report({'WARNING'},
                        "Some operations could not be performed (See Console for more info)")

            print("\n[Add Advanced  Objects]\nOperator: "
                  "objects_texture.add_scene\nError: {}".format(e))

            return {'CANCELLED'}

        return {'FINISHED'}



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see http://www.gnu.org/licenses/
#  or write to the Free Software Foundation, Inc., 51 Franklin Street,
#  Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

bl_info = {
    "name": "Copy2 Vertices, Edges or Faces",
    "author": "Eleanor Howick (elfnor.com)",
    "version": (0, 1, 1),
    "blender": (2, 71, 0),
    "location": "3D View > Object > Copy 2",
    "description": "Copy one object to the selected vertices, edges or faces of another object",
    "warning": "",
    "category": "Object"
}

import bpy
from bpy.types import Operator
from bpy.props import (
        BoolProperty,
        EnumProperty,
        FloatProperty,
        )
from mathutils import (
        Vector,
        Matrix,
        )


class Copy2(Operator):
    bl_idname = "mesh.copy2"
    bl_label = "Copy 2"
    bl_description = ("Copy Vertices, Edges or Faces to the Selected object\n"
                      "Needs an existing Active Mesh Object")
    bl_options = {"REGISTER", "UNDO"}

    obj_list = None

    def obj_list_cb(self, context):
        return Copy2.obj_list

    def sec_axes_list_cb(self, context):
        if self.priaxes == 'X':
            sec_list = [('Y', "Y", "Secondary axis Y"),
                        ('Z', "Z", "Secondary axis Z")]

        if self.priaxes == 'Y':
            sec_list = [('X', "X", "Secondary axis X"),
                        ('Z', "Z", "Secondary axis Z")]

        if self.priaxes == 'Z':
            sec_list = [('X', "X", "Secondary axis X"),
                        ('Y', "Y", "Secondary axis Y")]
        return sec_list

    copytype = EnumProperty(
            items=(('V', "Vertex",
                    "Paste the Copied Geometry to Vertices of the Active Object", 'VERTEXSEL', 0),
                   ('E', "Edge",
                    "Paste the Copied Geometry to Edges of the Active Object", 'EDGESEL', 1),
                   ('F', "Face",
                    "Paste the Copied Geometry to Faces of the Active Object", 'FACESEL', 2)),
            )
    copyfromobject = EnumProperty(
            name="Copy from",
            description="Copy an Object from the list",
            items=obj_list_cb
            )
    priaxes = EnumProperty(
            description="Primary axes used for Copied Object orientation",
            items=(('X', "X", "Along X"),
                   ('Y', "Y", "Along Y"),
                   ('Z', "Z", "Along Z")),
            )
    edgescale = BoolProperty(
            name="Scale to fill edge",
            default=False
            )
    secaxes = EnumProperty(
            name="Secondary Axis",
            description="Secondary axis used for Copied Object orientation",
            items=sec_axes_list_cb
            )
    scale = FloatProperty(
            name="Scale",
            default=1.0,
            min=0.0,
            )

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == "MESH"

    def draw(self, context):
        layout = self.layout

        layout.prop(self, "copyfromobject")
        layout.label("to:")
        layout.prop(self, "copytype", expand=True)
        layout.label("Primary axis:")
        layout.prop(self, "priaxes", expand=True)
        layout.label("Secondary axis:")
        layout.prop(self, "secaxes", expand=True)
        if self.copytype == "E":
            layout.prop(self, "edgescale")
            if self.edgescale:
                layout.prop(self, "scale")
        return

    def execute(self, context):
        copytoobject = context.active_object.name
        axes = self.priaxes + self.secaxes

        # check if there is a problem with the strings related to some chars
        copy_to_object = bpy.data.objects[copytoobject] if \
                         copytoobject in bpy.data.objects else None

        copy_from_object = bpy.data.objects[self.copyfromobject] if \
                           self.copyfromobject in bpy.data.objects else None

        if copy_to_object is None or copy_from_object is None:
            self.report({"WARNING"},
                        "There was a problem with retrieving Object data. Operation Cancelled")
            return {"CANCELLED"}
        try:
            copy_to_from(
                    context.scene,
                    copy_to_object,
                    copy_from_object,
                    self.copytype,
                    axes,
                    self.edgescale,
                    self.scale
                    )
        except Exception as e:
            self.report({"WARNING"},
                        "Copy2 could not be completed (Check the Console for more info)")
            print("\n[Add Advanced Objects]\nOperator: mesh.copy2\n{}\n".format(e))

            return {"CANCELLED"}

        return {"FINISHED"}

    def invoke(self, context, event):
        Copy2.obj_list = [(obj.name, obj.name, obj.name) for obj in bpy.data.objects]

        return {"FINISHED"}


def copy_to_from(scene, to_obj, from_obj, copymode, axes, edgescale, scale):
    if copymode == 'V':
        vertex_copy(scene, to_obj, from_obj, axes)

    if copymode == 'E':
        # don't pass edgescalling to object types that cannot be scaled
        if from_obj.type in ["CAMERA", "LAMP", "EMPTY", "ARMATURE", "SPEAKER", "META"]:
            edgescale = False
        edge_copy(scene, to_obj, from_obj, axes, edgescale, scale)

    if copymode == 'F':
        face_copy(scene, to_obj, from_obj, axes)


axes_dict = {'XY': (1, 2, 0),
             'XZ': (2, 1, 0),
             'YX': (0, 2, 1),
             'YZ': (2, 0, 1),
             'ZX': (0, 1, 2),
             'ZY': (1, 0, 2)}


def copyto(scene, source_obj, pos, xdir, zdir, axes, scale=None):
    """
    copy the source_obj to pos, so its primary axis points in zdir and its
    secondary axis points in xdir
    """
    copy_obj = source_obj.copy()
    scene.objects.link(copy_obj)

    xdir = xdir.normalized()
    zdir = zdir.normalized()
    # rotation first
    z_axis = zdir
    x_axis = xdir
    y_axis = z_axis.cross(x_axis)
    # use axes_dict to assign the axis as chosen in panel
    A, B, C = axes_dict[axes]
    rot_mat = Matrix()
    rot_mat[A].xyz = x_axis
    rot_mat[B].xyz = y_axis
    rot_mat[C].xyz = z_axis
    rot_mat.transpose()

    # rotate object
    copy_obj.matrix_world = rot_mat

    # move object into position
    copy_obj.location = pos

    # scale object
    if scale is not None:
        copy_obj.scale = scale

    return copy_obj


def vertex_copy(scene, obj, source_obj, axes):
    # vertex select mode
    sel_verts = []
    copy_list = []

    for v in obj.data.vertices:
        if v.select is True:
            sel_verts.append(v)

    # make a set for each vertex. The set contains all the connected vertices
    # use sets so the list is unique
    vert_con = [set() for i in range(len(obj.data.vertices))]
    for e in obj.data.edges:
        vert_con[e.vertices[0]].add(e.vertices[1])
        vert_con[e.vertices[1]].add(e.vertices[0])

    for v in sel_verts:
        pos = v.co * obj.matrix_world.transposed()
        xco = obj.data.vertices[list(vert_con[v.index])[0]].co * obj.matrix_world.transposed()

        zdir = (v.co + v.normal) * obj.matrix_world.transposed() - pos
        zdir = zdir.normalized()

        edir = pos - xco

        # edir is nor perpendicular to z dir
        # want xdir to be projection of edir onto plane through pos with direction zdir
        xdir = edir - edir.dot(zdir) * zdir
        xdir = -xdir.normalized()

        copy = copyto(scene, source_obj, pos, xdir, zdir, axes)
        copy_list.append(copy)

    # select all copied objects
    for copy in copy_list:
        copy.select = True
    obj.select = False


def edge_copy(scene, obj, source_obj, axes, es, scale):
    # edge select mode
    sel_edges = []
    copy_list = []

    for e in obj.data.edges:
        if e.select is True:
            sel_edges.append(e)

    for e in sel_edges:
        # pos is average of two edge vertexs
        v0 = obj.data.vertices[e.vertices[0]].co * obj.matrix_world.transposed()
        v1 = obj.data.vertices[e.vertices[1]].co * obj.matrix_world.transposed()
        pos = (v0 + v1) / 2
        # xdir is along edge
        xdir = v0 - v1
        xlen = xdir.magnitude
        xdir = xdir.normalized()
        # project each edge vertex normal onto plane normal to xdir
        vn0 = (obj.data.vertices[e.vertices[0]].co * obj.matrix_world.transposed() +
               obj.data.vertices[e.vertices[0]].normal) - v0
        vn1 = (obj.data.vertices[e.vertices[1]].co * obj.matrix_world.transposed() +
               obj.data.vertices[e.vertices[1]].normal) - v1
        vn0p = vn0 - vn0.dot(xdir) * xdir
        vn1p = vn1 - vn1.dot(xdir) * xdir
        # the mean of the two projected normals is the zdir
        zdir = vn0p + vn1p
        zdir = zdir.normalized()
        escale = None
        if es:
            escale = Vector([1.0, 1.0, 1.0])
            i = list('XYZ').index(axes[1])
            escale[i] = scale * xlen / source_obj.dimensions[i]

        copy = copyto(scene, source_obj, pos, xdir, zdir, axes, scale=escale)
        copy_list.append(copy)

    # select all copied objects
    for copy in copy_list:
        copy.select = True
    obj.select = False


def face_copy(scene, obj, source_obj, axes):
    # face select mode
    sel_faces = []
    copy_list = []

    for f in obj.data.polygons:
        if f.select is True:
            sel_faces.append(f)

    for f in sel_faces:
        fco = f.center * obj.matrix_world.transposed()
        # get first vertex corner of transformed object
        vco = obj.data.vertices[f.vertices[0]].co * obj.matrix_world.transposed()
        # get face normal of transformed object
        fn = (f.center + f.normal) * obj.matrix_world.transposed() - fco
        fn = fn.normalized()

        copy = copyto(scene, source_obj, fco, vco - fco, fn, axes)
        copy_list.append(copy)

    # select all copied objects
    for copy in copy_list:
        copy.select = True
    obj.select = False


def register():
    bpy.utils.register_class(Copy2)


def unregister():
    bpy.utils.unregister_class(Copy2)


if __name__ == "__main__":
    register()



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110 - 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

bl_info = {
    "name": "Topokit 2",
    "author": "dustractor",
    "version": (2, 0),
    "blender": (2, 60, 0),
    "location": "Edit mesh > Vertices/ Edges/ Faces menus",
    "description": "",
    "warning": "",
    "wiki_url": "",
    "category": "Mesh"}


import bpy
from bpy.types import Operator
# In between calls, this stores any data that is expensive or static,
# matched to the size of the mesh and the id of the operator that created it
cachedata = dict()
# tkey is moved to mesh_extra_tools\__init__.py register function


# just a mix-in for the operators...
class meshpoller:
    @classmethod
    def poll(self, context):
        try:
            assert context.active_object.type == "MESH"
        except:
            return False
        finally:
            return True


# BEGIN VERTICES SECTION

# This one works similarly to normal 'grow' (ctrl + NUMPAD_PLUS),
# except the original selection is not part of the result,
#
#   0--0--0         0--1--0
#   |  |  |         |  |  |
#   0--1--0  -->    1--0--1
#   |  |  |         |  |  |
#   0--0--0         0--1--0

class MESH_OT_vneighbors_edgewise(meshpoller, Operator):
    bl_idname = "mesh.v2v_by_edge"
    bl_label = "Neighbors by Edge"
    bl_description = ("Select neighbour vertices of a starting selected vertex\n"
                      "Similar to Grow Selection - apart from the\n"
                      "original selection is not part of the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))
        next_state = bytearray(meshkey[0])

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            vert_to_vert_map, prev_state = cachedata[meshkey]
        else:
            vert_to_vert_map = {i: {} for i in range(meshkey[0])}
            for a, b in mesh.edge_keys:
                vert_to_vert_map[a][b] = 1
                vert_to_vert_map[b][a] = 1
            obj.tkkey = meshkey
            prev_state = None

        if not prev_state:
            selected_vert_indices = filter(
                    lambda _: mesh.vertices[_].select,
                    range(len(mesh.vertices))
                    )
        else:
            selected_vert_indices = filter(
                    lambda _: mesh.vertices[_].select and not prev_state[_],
                    range(len(mesh.vertices))
                    )

        for v in selected_vert_indices:
            for neighbor_index in vert_to_vert_map[v]:
                next_state[neighbor_index] = True
        mesh.vertices.foreach_set("select", next_state)
        cachedata[meshkey] = (vert_to_vert_map, next_state)
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}


# This one is an alternate / counterpart to the previous.
# Think: diagonal opposite corners of a quad
# NOTE: does not apply to a triangle, since verts have no "opposite"
#
#   0--0--0     1--0--1
#   |  |  |     |  |  |
#   0--1--0 --> 0--0--0
#   |  |  |     |  |  |
#   0--0--0     1--0--1

class MESH_OT_vneighbors_facewise(meshpoller, Operator):
    bl_idname = "mesh.v2v_facewise"
    bl_label = "Neighbors by Face - Edge"
    bl_description = ("Select diagonal opposite vertices of neighbour quads\n"
                      "Does not work with triangles\n"
                      "The original selection is not part of the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))
        next_state = bytearray(meshkey[0])

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            vert_to_vert_map = cachedata[meshkey]
        else:
            vert_to_vert_map = {i: {} for i in range(meshkey[0])}
            for a, b in mesh.edge_keys:
                vert_to_vert_map[a][b] = 1
                vert_to_vert_map[b][a] = 1
            obj.tkkey = meshkey
        faces = filter(lambda face: (len(face.vertices) == 4) and
                       (face.select is False), mesh.polygons)
        for f in faces:
            has = False
            t = set()
            for v in f.vertices:
                if mesh.vertices[v].select:
                    has = True
                    t.update(vert_to_vert_map[v])
            if has:
                for v in f.vertices:
                    if not mesh.vertices[v].select:
                        if v not in t:
                            next_state[v] = 1
        mesh.vertices.foreach_set("select", next_state)
        cachedata[meshkey] = vert_to_vert_map
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}

# END VERTICES SECTION


# BEGIN EDGES SECTION
#   +--0--+--0--+--0--+          +--0--+--0--+--0--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     1     1     0
#   |     |     |     |          |     |     |     |
#   +--0--+--1--+--0--+   --->   +--0--+--0--+--0--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     1     1     0
#   |     |     |     |          |     |     |     |
#   +--0--+--0--+--0--+          +--0--+--0--+--0--+

class MESH_OT_eneighbors_shared_v_f(meshpoller, Operator):
    bl_idname = "mesh.e2e_evfe"
    bl_label = "Neighbors by Vert and Face"
    bl_description = ("Select edges that share the neighbour vertices and faces\n"
                      "of the starting selected edge\n"
                      "The original selection is not part of the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))
        state_mask = bytearray(meshkey[1])

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            edge_to_edges_dict = cachedata
        else:
            edge_key_to_index = {k: i for i, k in enumerate(mesh.edge_keys)}
            edge_to_edges_dict = {i: set() for i in range(len(mesh.edges))}
            for f in mesh.polygons:
                fed = [edge_key_to_index[k] for k in f.edge_keys]
                for k in f.edge_keys:
                    edge_to_edges_dict[edge_key_to_index[k]].update(fed)
            obj.tkkey = meshkey

        for e in filter(lambda _: mesh.edges[_].select, edge_to_edges_dict):
            k1 = set(mesh.edges[e].key)
            for n in edge_to_edges_dict[e]:
                k2 = set(mesh.edges[n].key)
                if not k1.isdisjoint(k2):
                    state_mask[n] = True

        for e in mesh.edges:
            e.select ^= state_mask[e.index]
        cachedata[meshkey] = edge_key_to_index
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}


#   +--0--+--0--+--0--+          +--0--+--0--+--0--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     1     1     0
#   |     |     |     |          |     |     |     |
#   +--0--+--1--+--0--+   --->   +--1--+--0--+--1--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     1     1     0
#   |     |     |     |          |     |     |     |
#   +--0--+--0--+--0--+          +--0--+--0--+--0--+

class MESH_OT_eneighbors_shared_v(meshpoller, Operator):
    bl_idname = "mesh.e2e_eve"
    bl_label = "Neighbors by Vert"
    bl_description = ("Select edges that share the neighbour vertices\n"
                     "of the starting selected edge\n"
                     "The original selection is not part of the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        bpy.ops.object.mode_set(mode="OBJECT")
        mesh = context.active_object.data
        state_mask = bytearray(len(mesh.edges))

        for e in mesh.edges:
            state_mask[e.index] = \
                    mesh.vertices[e.vertices[0]].select ^ mesh.vertices[e.vertices[1]].select
        mesh.edges.foreach_set('select', state_mask)
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}


#   +--0--+--0--+--0--+          +--0--+--1--+--0--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     1     1     0
#   |     |     |     |          |     |     |     |
#   +--0--+--1--+--0--+   --->   +--0--+--0--+--0--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     1     1     0
#   |     |     |     |          |     |     |     |
#   +--0--+--0--+--0--+          +--0--+--1--+--0--+

class MESH_OT_eneighbors_shared_f(meshpoller, Operator):
    bl_idname = "mesh.e2e_efe"
    bl_label = "Neighbors by Face"
    bl_description = ("Select edges of neighbour faces to the starting selected edge\n"
                     "The original selection is not part of the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            edge_to_edges_dict = cachedata
        else:
            edge_key_to_index = {k: i for i, k in enumerate(mesh.edge_keys)}
            edge_to_edges_dict = {i: set() for i in range(len(mesh.edges))}

            for f in mesh.polygons:
                fed = [edge_key_to_index[k] for k in f.edge_keys]
                for k in f.edge_keys:
                    edge_to_edges_dict[edge_key_to_index[k]].update(fed)

            obj.tkkey = meshkey
        state_mask, esel = (bytearray(meshkey[1]), bytearray(meshkey[1]))
        mesh.edges.foreach_get('select', esel)

        for e in filter(lambda _: mesh.edges[_].select, range(meshkey[1])):
            for n in edge_to_edges_dict[e]:
                state_mask[n] = 1

        for e in range(meshkey[1]):
            esel[e] ^= state_mask[e]
        mesh.edges.foreach_set('select', esel)
        cachedata[meshkey] = edge_to_edges_dict
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}


# Notice that on these next two, the original selection stays
#   +--0--+--0--+--0--+          +--0--+--1--+--0--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     0     0     0
#   |     |     |     |          |     |     |     |
#   +--0--+--1--+--0--+   --->   +--0--+--1--+--0--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     0     0     0
#   |     |     |     |          |     |     |     |
#   +--0--+--0--+--0--+          +--0--+--1--+--0--+

class MESH_OT_eneighbors_shared_f_notv(meshpoller, Operator):
    bl_idname = "mesh.e2e_efnve"
    bl_label = "Lateral Neighbors"
    bl_description = ("Select edges that are lateral neighbours\n"
                      "The original selection is included in the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))
        state_mask = bytearray(meshkey[1])

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            edge_to_face_map, edge_key_to_index = cachedata[meshkey]
        else:
            edge_key_to_index = {}
            edge_to_face_map = {i: set() for i in range(meshkey[1])}
            for i, k in enumerate(mesh. edge_keys):
                edge_key_to_index[k] = i

            for f in mesh.polygons:
                for k in f.edge_keys:
                    edge_to_face_map[edge_key_to_index[k]].add(f.index)
            obj.tkkey = meshkey
        selected_edge_indices = filter(lambda _: mesh.edges[_].select, range(meshkey[1]))

        for e in selected_edge_indices:
            for f in edge_to_face_map[e]:
                for k in mesh.polygons[f].edge_keys:
                    hasv_in = False
                    for v in mesh.edges[e].key:
                        if v in k:
                            hasv_in = True
                    if hasv_in:
                        continue
                    else:
                        state_mask[edge_key_to_index[k]] = True

        for e in filter(lambda _: state_mask[_], range(meshkey[1])):
            mesh.edges[e].select |= state_mask[e]
        cachedata[meshkey] = (edge_to_face_map, edge_key_to_index)
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}


#   +--0--+--0--+--0--+          +--0--+--0--+--0--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     0     0     0
#   |     |     |     |          |     |     |     |
#   +--0--+--1--+--0--+   --->   +--1--+--1--+--1--+
#   |     |     |     |          |     |     |     |
#   0     0     0     0          0     0     0     0
#   |     |     |     |          |     |     |     |
#   +--0--+--0--+--0--+          +--0--+--0--+--0--+

class MESH_OT_eneighbors_shared_v_notf(meshpoller, Operator):
    bl_idname = "mesh.e2e_evnfe"
    bl_label = "Longitudinal Edges"
    bl_description = ("Select Edges along the same longitude of the starting edge\n"
                      "The original selection is included in the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))
        state_mask = bytearray(meshkey[1])
        vstate = bytearray(meshkey[0])
        mesh.vertices.foreach_get('select', vstate)

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            edge_to_face_map, vert_to_vert_map, edge_key_to_index = cachedata[meshkey]
        else:
            edge_key_to_index = {}
            vert_to_vert_map = {i: set() for i in range(meshkey[0])}
            edge_to_face_map = {i: set() for i in range(meshkey[1])}

            for i, k in enumerate(mesh.edge_keys):
                edge_key_to_index[k] = i
                vert_to_vert_map[k[0]].add(k[1])
                vert_to_vert_map[k[1]].add(k[0])

            for f in mesh.polygons:
                for k in f.edge_keys:
                    edge_to_face_map[edge_key_to_index[k]].add(f.index)
            obj.tkkey = meshkey
        selected_edge_indices = filter(lambda _: mesh.edges[_].select, range(meshkey[1]))

        for e in selected_edge_indices:
            for v in mesh.edges[e].key:
                state_mask[v] ^= 1

            for f in edge_to_face_map[e]:
                for v in mesh.polygons[f].vertices:
                    vstate[v] = 1

        for v in filter(lambda _: state_mask[_], range(meshkey[1])):
            for n in vert_to_vert_map[v]:
                if not vstate[n] and (n != v):
                    mesh.edges[edge_key_to_index[(min(v, n), max(v, n))]].select = True
        cachedata[meshkey] = (edge_to_face_map, vert_to_vert_map, edge_key_to_index)
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}


# Deselects edges which are at the edge of a face-selection,
# causing selection to 'shrink in'
class MESH_OT_inner_edges(meshpoller, Operator):
    bl_idname = "mesh.ie"
    bl_label = "Inner Edge Selection"
    bl_description = ("Deselects edges which are at the border\n"
                      "of a starting face selection\n"
                      "causing the selection to shrink inwards")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))
        state_mask = bytearray(meshkey[1])

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            edge_to_face_map = cachedata[meshkey]
        else:
            edge_key_to_index = {k: i for i, k in enumerate(mesh.edge_keys)}
            edge_to_face_map = {i: set() for i in range(meshkey[1])}
            for f in mesh.polygons:
                for k in f.edge_keys:
                    edge_to_face_map[edge_key_to_index[k]].add(f.index)
            obj.tkkey = meshkey

        for e in filter(lambda _: mesh.edges[_].select, range(meshkey[1])):
            for f in edge_to_face_map[e]:
                if mesh.polygons[f].select:
                    state_mask[e] ^= 1

        for e in range(meshkey[1]):
            mesh.edges[e].select ^= state_mask[e]
        cachedata[meshkey] = edge_to_face_map
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}

# END EDGES SECTION


# BEGIN FACES SECTION

# here is another one which functions very similarly to the ctrl+NUMPAD_PLUS 'growth'
# but it deselects the original selection, of course.
# This would be your checkerboard-type growth.
#   [0][0][0]          [0][1][0]
#   [0][1][0]   --->   [1][0][1]
#   [0][0][0]          [0][1][0]

class MESH_OT_fneighbors_shared_e(meshpoller, Operator):
    bl_idname = "mesh.f2f_fef"
    bl_label = "Neighbor Faces sharing an Edge"
    bl_description = ("Selects faces that share an edge with the starting face selection\n"
                      "Similar to the Grow selection \n"
                      "The original selection is not part of the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            face_to_face_map = cachedata[meshkey]
        else:
            edge_key_to_index = {k: i for i, k in enumerate(mesh.edge_keys)}
            edge_to_face_map = {i: set() for i in range(meshkey[1])}
            for f in mesh.polygons:
                for k in f.edge_keys:
                    edge_to_face_map[edge_key_to_index[k]].add(f.index)
            face_to_face_map = {i: set() for i in range(meshkey[2])}
            for f in mesh.polygons:
                for k in f.edge_keys:
                    face_to_face_map[f.index].update(edge_to_face_map[edge_key_to_index[k]])
            obj.tkkey = meshkey
        mask_state = bytearray(meshkey[2])

        for f in filter(lambda _: mesh.polygons[_].select, range(meshkey[2])):
            for n in face_to_face_map[f]:
                mask_state[n] = True

        for f in range(meshkey[2]):
            mesh.polygons[f].select ^= mask_state[f]
        cachedata[meshkey] = face_to_face_map
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}


#   [0][0][0]          [1][0][1]
#   [0][1][0]   --->   [0][0][0]
#   [0][0][0]          [1][0][1]

class MESH_OT_fneighbors_shared_v_note(meshpoller, Operator):
    bl_idname = "mesh.f2f_fvnef"
    bl_label = "Neighbors by Vertex not Edge"
    bl_description = ("Select neighbour faces that share a vertex\n"
                      "with the starting selection\n"
                      "The original selection is not part of the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            edge_key_to_index = cachedata[meshkey]
        else:
            edge_key_to_index = {k: i for i, k in enumerate(mesh.edge_keys)}
            obj.tkkey = meshkey
        state_mask = bytearray(meshkey[2])
        face_verts = set()

        for f in filter(lambda _: mesh.polygons[_].select, range(meshkey[2])):
            face_verts.update(mesh.polygons[f].vertices)

        for f in filter(lambda _: not mesh.polygons[_].select, range(meshkey[2])):
            ct = 0
            for v in mesh.polygons[f].vertices:
                ct += (v in face_verts)
            if ct == 1:
                state_mask[f] = 1
        mesh.polygons.foreach_set('select', state_mask)
        cachedata[meshkey] = edge_key_to_index
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}


# https://en.wikipedia.org/wiki/Conway's_Game_of_Life
class MESH_OT_conway(meshpoller, Operator):
    bl_idname = "mesh.conway"
    bl_label = "Conway's Selection"
    bl_description = ("Select Faces with the Conway's game of life algorithm\n"
                      "Requires an initial Face selection\n"
                      "The edges of the original selection are included in the result")
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        global cachedata

        bpy.ops.object.mode_set(mode="OBJECT")
        obj = context.active_object
        mesh = obj.data
        meshkey = (len(mesh.vertices), len(mesh.edges), len(mesh.polygons), id(self))

        if (meshkey == obj.tkkey) and (meshkey in cachedata):
            vert_to_face_map = cachedata[meshkey]
        else:
            vert_to_face_map = {i: set() for i in range(meshkey[0])}
            for f in mesh.polygons:
                for v in f.vertices:
                    vert_to_face_map[v].add(f.index)
            obj.tkkey = meshkey
        sel = set()
        uns = set()
        F = {i: set() for i in range(meshkey[2])}

        for f in range(meshkey[2]):
            for v in mesh.polygons[f].vertices:
                for n in filter(lambda _: mesh.polygons[_].select and (_ != f), vert_to_face_map[v]):
                    F[f].add(n)

        for f in F:
            if len(F[f]) == 3:
                sel.add(f)
            elif len(F[f]) != 2:
                uns.add(f)

        for f in range(meshkey[2]):
            if f in sel:
                mesh.polygons[f].select = True
            if f in uns:
                mesh.polygons[f].select = False
        cachedata[meshkey] = vert_to_face_map
        bpy.ops.object.mode_set(mode="EDIT")

        return {"FINISHED"}


def register():
    bpy.utils.register_module(__name__)


def unregister():
    bpy.utils.unregister_module(__name__)


if __name__ == "__main__":
    register()



# ########################################################### #
#   An simple add-on to auto cut in two and mirror an object  #
#   Actualy partialy uncommented (see further version)        #
#   Author: Lapineige                                         #
#   License: GPL v3                                           #
# ########################################################### #

bl_info = {
    "name": "Auto Mirror",
    "description": "Super fast cutting and mirroring for Mesh objects",
    "author": "Lapineige",
    "version": (2, 4, 2),
    "blender": (2, 7, 1),
    "location": "View 3D > Toolbar > Tools tab > AutoMirror (panel)",
    "warning": "",
    "wiki_url": "https://wiki.blender.org/index.php/Extensions:2.6/"
                "Py/Scripts/Modeling/AutoMirror",
    "category": "Mesh"}


import bpy
from bpy.props import (
        BoolProperty,
        EnumProperty,
        FloatProperty,
        PointerProperty,
        )
from bpy.types import (
        Operator,
        Panel,
        PropertyGroup,
        )
from mathutils import Vector


# Operators
class AlignVertices(Operator):
    bl_idname = "object.align_vertices"
    bl_label = "Align Vertices on an Axis"
    bl_description = ("Align Vertices on an Axis\n"
                      "Needs an Active Mesh Object")

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == "MESH"

    def execute(self, context):
        auto_m = context.scene.auto_mirror
        bpy.ops.object.mode_set(mode='OBJECT')

        x1, y1, z1 = bpy.context.scene.cursor_location
        bpy.ops.view3d.snap_cursor_to_selected()

        x2, y2, z2 = bpy.context.scene.cursor_location

        bpy.context.scene.cursor_location[0], \
        bpy.context.scene.cursor_location[1], \
        bpy.context.scene.cursor_location[2] = 0, 0, 0

        # Vertices coordinate to 0 (local coordinate, so on the origin)
        for vert in bpy.context.object.data.vertices:
            if vert.select:
                if auto_m.axis == 'x':
                    axis = 0
                elif auto_m.axis == 'y':
                    axis = 1
                elif auto_m.axis == 'z':
                    axis = 2
                vert.co[axis] = 0

        bpy.context.scene.cursor_location = x2, y2, z2
        bpy.ops.object.origin_set(type='ORIGIN_CURSOR')

        bpy.context.scene.cursor_location = x1, y1, z1
        bpy.ops.object.mode_set(mode='EDIT')

        return {'FINISHED'}


class AutoMirror(Operator):
    bl_idname = "object.automirror"
    bl_label = "AutoMirror"
    bl_description = ("Automatically cut an object along an axis\n"
                      "Needs an Active Mesh Object")
    bl_options = {'REGISTER'}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == "MESH"

    def get_local_axis_vector(self, context, X, Y, Z, orientation):
        loc = context.object.location
        bpy.ops.object.mode_set(mode="OBJECT")  # Needed to avoid to translate vertices

        v1 = Vector((loc[0], loc[1], loc[2]))
        bpy.ops.transform.translate(
                value=(X * orientation, Y * orientation, Z * orientation),
                constraint_axis=((X == 1), (Y == 1), (Z == 1)),
                constraint_orientation='LOCAL'
                )
        v2 = Vector((loc[0], loc[1], loc[2]))
        bpy.ops.transform.translate(
                value=(-X * orientation, -Y * orientation, -Z * orientation),
                constraint_axis=((X == 1), (Y == 1), (Z == 1)),
                constraint_orientation='LOCAL'
                )

        bpy.ops.object.mode_set(mode="EDIT")
        return v2 - v1

    def execute(self, context):
        auto_m = context.scene.auto_mirror

        X, Y, Z = 0, 0, 0
        if auto_m.axis == 'x':
            X = 1
        elif auto_m.axis == 'y':
            Y = 1
        elif auto_m.axis == 'z':
            Z = 1

        current_mode = bpy.context.object.mode    # Save the current mode

        if bpy.context.object.mode != "EDIT":
            bpy.ops.object.mode_set(mode="EDIT")  # Go to edit mode

        bpy.ops.mesh.select_all(action='SELECT')  # Select all the vertices
        if auto_m.orientation == 'positive':
            orientation = 1
        else:
            orientation = -1
        cut_normal = self.get_local_axis_vector(context, X, Y, Z, orientation)

        # Cut the mesh
        bpy.ops.mesh.bisect(
                plane_co=(
                    bpy.context.object.location[0],
                    bpy.context.object.location[1],
                    bpy.context.object.location[2]
                    ),
                plane_no=cut_normal,
                use_fill=False,
                clear_inner=auto_m.cut,
                clear_outer=0,
                threshold=auto_m.threshold
                )

        # Use to align the vertices on the origin, needed by the "threshold"
        bpy.ops.object.align_vertices()

        if not auto_m.toggle_edit:
            bpy.ops.object.mode_set(mode=current_mode)  # Reload previous mode

        if auto_m.cut:
            bpy.ops.object.modifier_add(type='MIRROR')  # Add a mirror modifier
            bpy.context.object.modifiers[-1].use_x = X  # Choose the axis to use, based on the cut's axis
            bpy.context.object.modifiers[-1].use_y = Y
            bpy.context.object.modifiers[-1].use_z = Z
            bpy.context.object.modifiers[-1].use_clip = auto_m.use_clip
            bpy.context.object.modifiers[-1].show_on_cage = auto_m.show_on_cage

            if auto_m.apply_mirror:
                bpy.ops.object.mode_set(mode='OBJECT')
                bpy.ops.object.modifier_apply(
                        apply_as='DATA',
                        modifier=bpy.context.object.modifiers[-1].name
                        )
                if auto_m.toggle_edit:
                    bpy.ops.object.mode_set(mode='EDIT')
                else:
                    bpy.ops.object.mode_set(mode=current_mode)

        return {'FINISHED'}


# Panel
class BisectMirror(Panel):
    bl_label = "Auto Mirror"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'TOOLS'
    bl_category = "Tools"
    bl_options = {"DEFAULT_CLOSED"}

    def draw(self, context):
        layout = self.layout
        auto_m = context.scene.auto_mirror
        obj = context.active_object

        if obj and obj.type == 'MESH':
            layout.operator("object.automirror", icon="MOD_MIRROR")
            layout.label("Options:")
            layout.prop(auto_m, "axis", text="Mirror Axis", expand=True)
            layout.prop(auto_m, "orientation", text="Orientation")
            layout.prop(auto_m, "threshold", text="Threshold")
            layout.prop(auto_m, "toggle_edit", text="Toggle Edit")
            layout.prop(auto_m, "cut", text="Cut and Mirror", toggle=True, icon="MOD_REMESH")

            if auto_m.cut:
                col = layout.column(align=True)
                row = col.row(align=True)
                row.prop(auto_m, "use_clip", text="Use Clip", toggle=True)
                row.prop(auto_m, "show_on_cage", text="Editable", toggle=True)
                col.prop(auto_m, "apply_mirror", text="Apply Mirror", toggle=True)
        else:
            layout.label(icon="INFO", text="No Mesh selected")


class AutoMirrorProperties(PropertyGroup):
    axis = EnumProperty(
            name="Axis",
            items=[
                ("x", "X", "", 1),
                ("y", "Y", "", 2),
                ("z", "Z", "", 3)
                ],
            description="Axis used by the mirror modifier"
            )
    orientation = EnumProperty(
            name="Orientation",
            items=[
                ("positive", "Positive", "", 1),
                ("negative", "Negative", "", 2)
                ],
            description="Choose the side along the axis of the editable part (+/- coordinates)"
            )
    threshold = FloatProperty(
            default=0.001,
            min=0.001,
            description="Vertices closer than this distance are merged on the loopcut"
            )
    toggle_edit = BoolProperty(
            name="Toggle Edit Mode",
            default=True,
            description="If not in Edit mode, change mode to it"
            )
    cut = BoolProperty(
            name="Cut",
            default=True,
            description="If enabled, cut the mesh in two parts and mirror it\n"
                        "If not, just make a loopcut"
            )
    clipping = BoolProperty(
            default=True
            )
    use_clip = BoolProperty(
            default=True,
            description="Use clipping for the mirror modifier"
            )
    show_on_cage = BoolProperty(
            default=True,
            description="Enable editing the cage (it's the classical modifier's option)"
            )
    apply_mirror = BoolProperty(
            description="Apply the mirror modifier (useful to symmetrise the mesh)"
            )


def register():
    bpy.utils.register_class(BisectMirror)
    bpy.utils.register_class(AutoMirror)
    bpy.utils.register_class(AlignVertices)
    bpy.utils.register_class(AutoMirrorProperties)
    bpy.types.Scene.auto_mirror = PointerProperty(
                                        type=AutoMirrorProperties
                                        )


def unregister():
    bpy.utils.unregister_class(BisectMirror)
    bpy.utils.unregister_class(AutoMirror)
    bpy.utils.unregister_class(AlignVertices)
    bpy.utils.unregister_class(AutoMirrorProperties)
    del bpy.types.Scene.auto_mirror


if __name__ == "__main__":
    register()



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END GPL LICENSE BLOCK #####

# Contact for more information about the Addon:
# Email:    germano.costa@ig.com.br
# Twitter:  wii_mano @mano_wii

bl_info = {
    "name": "Extrude and Reshape",
    "author": "Germano Cavalcante",
    "version": (0, 8, 1),
    "blender": (2, 76, 5),
    "location": "View3D > TOOLS > Tools > Mesh Tools > Add: > Extrude Menu (Alt + E)",
    "description": "Extrude face and merge edge intersections "
                   "between the mesh and the new edges",
    "wiki_url": "http://blenderartists.org/forum/"
                "showthread.php?376618-Addon-Push-Pull-Face",
    "category": "Mesh"}

import bpy
import bmesh
from mathutils.geometry import intersect_line_line
from bpy.types import Operator


class BVHco():
    i = 0
    c1x = 0.0
    c1y = 0.0
    c1z = 0.0
    c2x = 0.0
    c2y = 0.0
    c2z = 0.0


def edges_BVH_overlap(bm, edges, epsilon=0.0001):
    bco = set()
    for e in edges:
        bvh = BVHco()
        bvh.i = e.index
        b1 = e.verts[0]
        b2 = e.verts[1]
        co1 = b1.co.x
        co2 = b2.co.x
        if co1 <= co2:
            bvh.c1x = co1 - epsilon
            bvh.c2x = co2 + epsilon
        else:
            bvh.c1x = co2 - epsilon
            bvh.c2x = co1 + epsilon
        co1 = b1.co.y
        co2 = b2.co.y
        if co1 <= co2:
            bvh.c1y = co1 - epsilon
            bvh.c2y = co2 + epsilon
        else:
            bvh.c1y = co2 - epsilon
            bvh.c2y = co1 + epsilon
        co1 = b1.co.z
        co2 = b2.co.z
        if co1 <= co2:
            bvh.c1z = co1 - epsilon
            bvh.c2z = co2 + epsilon
        else:
            bvh.c1z = co2 - epsilon
            bvh.c2z = co1 + epsilon
        bco.add(bvh)
    del edges
    overlap = {}
    oget = overlap.get
    for e1 in bm.edges:
        by = bz = True
        a1 = e1.verts[0]
        a2 = e1.verts[1]
        c1x = a1.co.x
        c2x = a2.co.x
        if c1x > c2x:
            tm = c1x
            c1x = c2x
            c2x = tm
        for bvh in bco:
            if c1x <= bvh.c2x and c2x >= bvh.c1x:
                if by:
                    by = False
                    c1y = a1.co.y
                    c2y = a2.co.y
                    if c1y > c2y:
                        tm = c1y
                        c1y = c2y
                        c2y = tm
                if c1y <= bvh.c2y and c2y >= bvh.c1y:
                    if bz:
                        bz = False
                        c1z = a1.co.z
                        c2z = a2.co.z
                        if c1z > c2z:
                            tm = c1z
                            c1z = c2z
                            c2z = tm
                    if c1z <= bvh.c2z and c2z >= bvh.c1z:
                        e2 = bm.edges[bvh.i]
                        if e1 != e2:
                            overlap[e1] = oget(e1, set()).union({e2})
    return overlap


def intersect_edges_edges(overlap, precision=4):
    epsilon = .1**precision
    fpre_min = -epsilon
    fpre_max = 1 + epsilon
    splits = {}
    sp_get = splits.get
    new_edges1 = set()
    new_edges2 = set()
    targetmap = {}
    for edg1 in overlap:
        # print("***", ed1.index, "***")
        for edg2 in overlap[edg1]:
            a1 = edg1.verts[0]
            a2 = edg1.verts[1]
            b1 = edg2.verts[0]
            b2 = edg2.verts[1]

            # test if are linked
            if a1 in {b1, b2} or a2 in {b1, b2}:
                # print('linked')
                continue

            aco1, aco2 = a1.co, a2.co
            bco1, bco2 = b1.co, b2.co
            tp = intersect_line_line(aco1, aco2, bco1, bco2)
            if tp:
                p1, p2 = tp
                if (p1 - p2).to_tuple(precision) == (0, 0, 0):
                    v = aco2 - aco1
                    f = p1 - aco1
                    x, y, z = abs(v.x), abs(v.y), abs(v.z)
                    max1 = 0 if x >= y and x >= z else\
                           1 if y >= x and y >= z else 2
                    fac1 = f[max1] / v[max1]

                    v = bco2 - bco1
                    f = p2 - bco1
                    x, y, z = abs(v.x), abs(v.y), abs(v.z)
                    max2 = 0 if x >= y and x >= z else\
                           1 if y >= x and y >= z else 2
                    fac2 = f[max2] / v[max2]

                    if fpre_min <= fac1 <= fpre_max:
                        # print(edg1.index, 'can intersect', edg2.index)
                        ed1 = edg1

                    elif edg1 in splits:
                        for ed1 in splits[edg1]:
                            a1 = ed1.verts[0]
                            a2 = ed1.verts[1]

                            vco1 = a1.co
                            vco2 = a2.co

                            v = vco2 - vco1
                            f = p1 - vco1
                            fac1 = f[max1] / v[max1]
                            if fpre_min <= fac1 <= fpre_max:
                                # print(e.index, 'can intersect', edg2.index)
                                break
                        else:
                            # print(edg1.index, 'really does not intersect', edg2.index)
                            continue
                    else:
                        # print(edg1.index, 'not intersect', edg2.index)
                        continue

                    if fpre_min <= fac2 <= fpre_max:
                        # print(ed1.index, 'actually intersect', edg2.index)
                        ed2 = edg2

                    elif edg2 in splits:
                        for ed2 in splits[edg2]:
                            b1 = ed2.verts[0]
                            b2 = ed2.verts[1]

                            vco1 = b1.co
                            vco2 = b2.co

                            v = vco2 - vco1
                            f = p2 - vco1
                            fac2 = f[max2] / v[max2]
                            if fpre_min <= fac2 <= fpre_max:
                                # print(ed1.index, 'actually intersect', e.index)
                                break
                        else:
                            # print(ed1.index, 'really does not intersect', ed2.index)
                            continue
                    else:
                        # print(ed1.index, 'not intersect', edg2.index)
                        continue

                    new_edges1.add(ed1)
                    new_edges2.add(ed2)

                    if abs(fac1) <= epsilon:
                        nv1 = a1
                    elif fac1 + epsilon >= 1:
                        nv1 = a2
                    else:
                        ne1, nv1 = bmesh.utils.edge_split(ed1, a1, fac1)
                        new_edges1.add(ne1)
                        splits[edg1] = sp_get(edg1, set()).union({ne1})

                    if abs(fac2) <= epsilon:
                        nv2 = b1
                    elif fac2 + epsilon >= 1:
                        nv2 = b2
                    else:
                        ne2, nv2 = bmesh.utils.edge_split(ed2, b1, fac2)
                        new_edges2.add(ne2)
                        splits[edg2] = sp_get(edg2, set()).union({ne2})

                    if nv1 != nv2:  # necessary?
                        targetmap[nv1] = nv2

    return new_edges1, new_edges2, targetmap


class Extrude_and_Reshape(Operator):
    bl_idname = "mesh.extrude_reshape"
    bl_label = "Extrude and Reshape"
    bl_description = "Push and pull face entities to sculpt 3d models"
    bl_options = {'REGISTER', 'GRAB_CURSOR', 'BLOCKING'}

    @classmethod
    def poll(cls, context):
        return context.mode is not 'EDIT_MESH'

    def modal(self, context, event):
        if self.confirm:
            sface = self.bm.faces.active
            if not sface:
                for face in self.bm.faces:
                    if face.select is True:
                        sface = face
                        break
                else:
                    return {'FINISHED'}
            # edges to intersect
            edges = set()
            [[edges.add(ed) for ed in v.link_edges] for v in sface.verts]

            overlap = edges_BVH_overlap(self.bm, edges, epsilon=0.0001)
            overlap = {k: v for k, v in overlap.items() if k not in edges}  # remove repetition
            """
            print([e.index for e in edges])
            for a, b in overlap.items():
                print(a.index, [e.index for e in b])
            """
            new_edges1, new_edges2, targetmap = intersect_edges_edges(overlap)
            pos_weld = set()
            for e in new_edges1:
                v1, v2 = e.verts
                if v1 in targetmap and v2 in targetmap:
                    pos_weld.add((targetmap[v1], targetmap[v2]))
            if targetmap:
                bmesh.ops.weld_verts(self.bm, targetmap=targetmap)
            """
            print([e.is_valid for e in new_edges1])
            print([e.is_valid for e in new_edges2])
            sp_faces1 = set()
            """
            for e in pos_weld:
                v1, v2 = e
                lf1 = set(v1.link_faces)
                lf2 = set(v2.link_faces)
                rlfe = lf1.intersection(lf2)
                for f in rlfe:
                    try:
                        nf = bmesh.utils.face_split(f, v1, v2)
                        # sp_faces1.update({f, nf[0]})
                    except:
                        pass

            # sp_faces2 = set()
            for e in new_edges2:
                lfe = set(e.link_faces)
                v1, v2 = e.verts
                lf1 = set(v1.link_faces)
                lf2 = set(v2.link_faces)
                rlfe = lf1.intersection(lf2)
                for f in rlfe.difference(lfe):
                    nf = bmesh.utils.face_split(f, v1, v2)
                    # sp_faces2.update({f, nf[0]})

            bmesh.update_edit_mesh(self.mesh, tessface=True, destructive=True)
            return {'FINISHED'}
        if self.cancel:
            return {'FINISHED'}
        self.cancel = event.type in {'ESC', 'NDOF_BUTTON_ESC'}
        self.confirm = event.type in {'LEFTMOUSE', 'RET', 'NUMPAD_ENTER'}
        return {'PASS_THROUGH'}

    def execute(self, context):
        self.mesh = context.object.data
        self.bm = bmesh.from_edit_mesh(self.mesh)
        try:
            selection = self.bm.select_history[-1]
        except:
            for face in self.bm.faces:
                if face.select is True:
                    selection = face
                    break
            else:
                return {'FINISHED'}
        if not isinstance(selection, bmesh.types.BMFace):
            bpy.ops.mesh.extrude_region_move('INVOKE_DEFAULT')
            return {'FINISHED'}
        else:
            face = selection
            # face.select = False
            bpy.ops.mesh.select_all(action='DESELECT')
            geom = []
            for edge in face.edges:
                if abs(edge.calc_face_angle(0) - 1.5707963267948966) < 0.01:  # self.angle_tolerance:
                    geom.append(edge)

            ret_dict = bmesh.ops.extrude_discrete_faces(self.bm, faces=[face])

            for face in ret_dict['faces']:
                self.bm.faces.active = face
                face.select = True
                sface = face
            dfaces = bmesh.ops.dissolve_edges(
                        self.bm, edges=geom, use_verts=True, use_face_split=False
                        )
            bmesh.update_edit_mesh(self.mesh, tessface=True, destructive=True)
            bpy.ops.transform.translate(
                        'INVOKE_DEFAULT', constraint_axis=(False, False, True),
                        constraint_orientation='NORMAL', release_confirm=True
                        )

        context.window_manager.modal_handler_add(self)

        self.cancel = False
        self.confirm = False
        return {'RUNNING_MODAL'}


def operator_draw(self, context):
    layout = self.layout
    col = layout.column(align=True)
    col.operator("mesh.extrude_reshape", text="Extrude and Reshape")


def register():
    bpy.utils.register_class(Extrude_and_Reshape)
    bpy.types.VIEW3D_MT_edit_mesh_extrude.append(operator_draw)


def unregister():
    bpy.types.VIEW3D_MT_edit_mesh_extrude.remove(operator_draw)
    bpy.utils.unregister_class(Extrude_and_Reshape)


if __name__ == "__main__":
    register()
